from __future__ import annotations

import importlib.metadata
import logging
import os
import textwrap
import time
from datetime import datetime
from pathlib import Path
from subprocess import run
from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    from typing import Any, Dict, List, Optional, Type

logger = logging.getLogger(__name__)


def get_Tasker(system: str) -> Type[Tasker]:
    """Function to get a Tasker object for a given system.

    Args:
        system : Shell the Tasker will use. 'slurm_cori' to use
            slurm scripts on cori, slurm_perlmutter' to use slurm scripts
            on perlmutter, 'bash' to run it in login nodes or computer shell.
    """
    if system == "slurm":
        from warnings import warn

        warn(
            "slurm system option is deprecated,"
            "use slurm_cori or slurm_perlmutter instead",
            DeprecationWarning,
        )
        return SlurmCoriTasker
    elif system == "slurm_cori":
        return SlurmCoriTasker
    elif system == "slurm_perlmutter":
        return SlurmPerlmutterTasker
    elif system == "bash":
        return BashTasker
    else:
        raise ValueError(
            f"System not valid: {system} "
            "allowed options are slurm_cori, slurm_perlmutter, "
            "bash."
        )


class Tasker:
    """Object to write and run jobs.

    Attributes:
        slurm_header_args (dict): Header options to write if slurm tasker is selected.
            Use a dictionary with the format {'option_name': 'option_value'}
        command (str): Command to be run in the job.
        command_args (str): Arguments to the command.
        packages (List[str]): Packages to list versions from.
        precommand (str): Instruction to run right before the command.
        environment (str): Conda/python environment to load before running the command.
        environmental_variables (dict): Environmental variables to set before running the job.
            Format: {'environmental_variable': 'value'}.
        srun_options (str): If slurm tasker selected. Options for the srun command.
        run_file (Path): Location of the job file.
        wait_for (Tasker or str): If slurm tasker selected. Tasker to wait for before
            running (if Tasker) or jobid of the task to wait (if str).
        in_files: Input files that must exists or contain a jobid in order for the
            job to be launched.
        out_files: Out file that will be write by the job (to add jobid if available).
    """

    default_srun_options: Dict = dict()
    default_header: Dict = dict()

    def __init__(
        self,
        command: str,
        command_args: Dict,
        slurm_header_args: Dict,
        environment: str,
        run_file: Path | str,
        jobid_log_file: Path | str,
        packages: List[str] = [],
        wait_for: Optional[
            ChainedTasker | Tasker | List[Type[Tasker]] | int | List[int]
        ] = None,
        environmental_variables: Dict = dict(),
        srun_options: Dict = dict(),
        in_files: List[Path] | List[str] = list[Path](),
        out_files: List[Path | str] = [],
        precommand: str = "",
    ):
        """
        Args:
            command (str): Command to be run in the job.
            command_args (str): Arguments to the command.
            packages (List[str]): Packages to list versions from.
            slurm_header_args (dict): Header options to write if slurm tasker is selected. Use a dictionary with the format {'option_name': 'option_value'}.
            srun_options (str): If slurm tasker selected. Options for the srun command.
            environment (str): Conda/python environment to load before running the command.
            run_file (Path): Location of the job file.
            wait_for (Tasker or int, optional): In NERSC, wait for a given job to finish before running the current one. Could be a  Tasker object or a slurm jobid (int). (Default: None, won't wait for anything).
            environmental_variables (dict, optional): Environmental variables to set before running the job. Format: {'environmental_variable': 'value'}. Default: No environmental variables defined.
            jobid_log_file (Path): Location of log file where to include jobids of runs.
            in_files: Input files that must exists or contain a jobid in order for the
            job to be launched.
            out_files: Out files that will be write by the job (to add jobid if available).
        """
        self.jobid: Optional[int] = None

        if "OMP_NUM_THREADS" in slurm_header_args:
            if slurm_header_args["OMP_NUM_THREADS"] not in ("", None):
                self.OMP_threads = slurm_header_args["OMP_NUM_THREADS"]
            else:
                self.OMP_threads = None

            del slurm_header_args["OMP_NUM_THREADS"]
        else:
            self.OMP_threads = None

        self.slurm_header_args = {**self.default_header, **slurm_header_args}

        if isinstance(self.slurm_header_args.get("time", ""), int):
            print(self.slurm_header_args["time"])
            logger.warning(
                "Detected int value in field time inside slurm args. "
                "Be sure to quote time values in config file if the "
                "format is hours:minutes:seconds. "
                "If you sent an int as minutes, ignore this warning."
            )
        self.command = command
        self.command_args = command_args
        self.packages = packages
        self.environment = environment
        self.environmental_variables = environmental_variables
        self.srun_options = {**self.default_srun_options, **srun_options}
        self.run_file = Path(run_file)
        self.wait_for = wait_for
        self.jobid_log_file = jobid_log_file
        self.in_files = in_files
        self.out_files = out_files
        self.precommand = precommand

    def get_wait_for_ids(self) -> None:
        """Method to standardise wait_for Taskers or ids, in such a way that can be easily used afterwards."""
        self.wait_for = list(np.array(self.wait_for).reshape(-1))
        self.wait_for_ids = []

        for x in self.wait_for:
            if isinstance(x, (int, np.integer)):
                self.wait_for_ids.append(x)
            elif isinstance(x, Tasker):
                try:
                    self.wait_for_ids.append(x.jobid)
                except AttributeError as e:
                    raise Exception(
                        f"jobid not defined for some of the wait_for. Have you send "
                        "any script with each of the wait_for objects?"
                    ).with_traceback(e.__traceback__)
            elif isinstance(x, type(None)):
                pass
            else:
                raise ValueError("Unrecognized wait_for object: ", x)

        for file in self.in_files:
            file = Path(file)
            if not file.is_file():
                raise FileNotFoundError("Input file for run not found", str(file))

            size = file.stat().st_size
            if size > 1 and size < 20:
                jobid = int(Path(file).read_text().splitlines()[0])
                status = self.get_jobid_status(jobid)

                if status != "COMPLETED":
                    self.wait_for_ids.append(jobid)

    def _make_command(self) -> str:
        """Method to generate a command with args

        Returns:
            str: Command to write in job run.
        """
        args_list = []
        for key, value in self.command_args.items():
            if key != "":
                if key.startswith("-"):
                    args_list.append(f"{key} {value}")
                else:
                    args_list.append(f"--{key} {value}")
            else:
                args_list.append(str(value))

        args = " ".join(args_list)
        return f'command="{self.command} {args}"'

    def _make_body(self) -> str:
        """Method to generate the body of the job script.

        Return:
            str: Body of the job script."""
        header = self._make_header()
        env_opts = self._make_env_opts()
        version_control = self._make_version_control()
        command = self._make_command()
        run_command = self._make_run_command()

        return "\n".join(
            [
                header,
                env_opts,
                version_control,
                command,
                "date",
                run_command,
                "date",
                "",
            ]
        )

    def write_job(self) -> None:
        """Method to write job script into file."""
        with open(self.run_file, "w") as f:
            f.write(self._make_body())

    def write_jobid(self) -> None:
        """Method to write jobid into log file."""
        with open(self.jobid_log_file, "a") as file:
            file.write(str(self.run_file.name) + " " + str(self.jobid) + "\n")

        for out_file in self.out_files:
            with open(out_file, "w") as file:
                file.write(str(self.jobid))

    def send_job(self) -> None:
        raise ValueError(
            "Tasker class has no send_job defined, use child classes instead."
        )

    def _make_header(self) -> str:
        raise ValueError(
            "Tasker class has no _make_header defined, use child classes instead."
        )

    def _make_env_opts(self) -> str:
        raise ValueError(
            "Tasker class has no _make_env_opts defined, use child classes instead."
        )

    def _make_version_control(self) -> str:
        text = "\necho used picca_bookkeeper version: "
        text += importlib.metadata.version("picca_bookkeeper")

        for package in self.packages:
            text += f'\necho using {package} version: $(python -c "import importlib.metadata; '
            text += f"print(importlib.metadata.version('{package}'))\")"

        return text + "\necho -e '\\n'\n"

    def _make_run_command(self) -> str:
        raise ValueError(
            "Tasker class has no _make_run_command defined, use child classes instead."
        )

    @staticmethod
    def get_jobid_status(jobid: int) -> str:
        """
        Method to return the status of a given jobid in SLURM systems.

        Args:
            jobid: Identificator of the job to obtain the status for.
        """
        tries = 0
        while tries < 10:
            sbatch_process = run(
                f"sacct -j {jobid} -o State --parsable2 -n",
                shell=True,
                capture_output=True,
            )

            try:
                return sbatch_process.stdout.decode("utf-8").splitlines()[-1]
            except:
                logger.debug(
                    f"Retrieving status for jobid {jobid} failed. Retrying in 2 seconds..."
                )
                time.sleep(2)

        logger.info(f"Retrieving status failed. Assuming job failed.")
        return "FAILED"


class SlurmTasker(Tasker):
    """Object to write and run jobs.

    Attributes:
        slurm_header_args (dict): Header options to write. Use a dictionary with the format {'option_name': 'option_value'}
        command (str): Command to be run in the job.
        command_args (str): Arguments to the command.
        environment (str): Conda/python environment to load before running the command.
        environmental_variables (dict): Environmental variables to set before running the job. Format: {'environmental_variable': 'value'}.
        srun_options (str): Options for the srun command.
        run_file (Path): Location of the job file.
        wait_for (Tasker or str): Tasker to wait for before running (if Tasker) or jobid of the task to wait (if str).
    """

    default_srun_options = {
        "nodes": 1,  # N
        "ntasks": 1,  # n
    }

    default_header = {
        "qos": "regular",
        "nodes": 1,
        "time": "00:30:00",
        "cpus-per-task": "256",
    }

    def __init__(self, *args: Any, **kwargs: Any):
        """
        Args:
            command (str): Command to be run in the job.
            command_args (str): Arguments to the command.
            slurm_header_args (dict): Header options to write if slurm tasker is selected. Use a dictionary with the format {'option_name': 'option_value'}.
            srun_options (str): If slurm tasker selected. Options for the srun command.
            environment (str): Conda/python environment to load before running the command.
            run_file (Path): Location of the job file.
            wait_for (Tasker or int, optional): In NERSC, wait for a given job to finish before running the current one. Could be a  Tasker object or a slurm jobid (int). (Default: None, won't wait for anything).
            environmental_variables (dict, optional): Environmental variables to set before running the job. Format: {'environmental_variable': 'value'}. Default: No environmental variables defined.
        """
        super().__init__(*args, **kwargs)

    def _make_header(self) -> str:
        """Method to generate a slurm header given the slurm_header_args attribute.

        Returns:
            str: Slurm header.
        """
        header = "#!/bin/bash -l\n\n"
        header += "\n".join(
            [
                f"#SBATCH --{key} {value}"
                for key, value in self.slurm_header_args.items()
            ]
        )
        return header

    def _make_env_opts(self) -> str:
        """Method to generate environmental options.

        Returns:
            str: environmental options."""
        if self.OMP_threads is not None:
            text = f"export OMP_NUM_THREADS={self.OMP_threads}\n"
        else:
            text = ""

        if self.environment.endswith(".sh"):
            activate = "source "
        else:
            activate = "conda activate "

        text += textwrap.dedent(
            f"""
module load python
{activate}{self.environment}
umask 0002

"""
        )
        for key, value in self.environmental_variables.items():
            text += f"export {key}={value}\n"

        text += self.precommand + "\n"

        return text

    def _make_run_command(self) -> str:
        """Method to generate the srun command.

        Returns:
            str: srun command."""
        args = " ".join(
            [f"--{key} {value}" for key, value in self.srun_options.items()]
        )
        return f"srun {args} $command\n"

    def send_job(self) -> None:
        """Method to send job to slurm queue. Setting the wait_for variables beforehand."""
        # Changing dir into the file position to
        # avoid slurm files being written in unwanted places
        _ = Path(".").resolve()
        os.chdir(self.run_file.parent)

        if self.wait_for is None and len(self.in_files) == 0:
            wait_for_str = ""
        else:
            self.get_wait_for_ids()

            if len(self.wait_for_ids) == 0:
                wait_for_str = ""
            else:
                wait_for_str = f"--dependency=afterok:"
                wait_for_str += ",afterok:".join(map(str, self.wait_for_ids))

        self.sbatch_process = run(
            f"sbatch --parsable {wait_for_str} {self.run_file}",
            shell=True,
            capture_output=True,
        )
        os.chdir(_)

        if self.sbatch_process.returncode == 0:
            self.jobid = int(self.sbatch_process.stdout.decode("utf-8"))
            self.write_jobid()
        else:
            raise ValueError(
                f'Submitting job failed. {self.sbatch_process.stderr.decode("utf-8")}'
            )


class SlurmCoriTasker(SlurmTasker):
    default_header = {
        "qos": "regular",
        "nodes": 1,
        "time": "00:30:00",
        "constraint": "haswell",
        "account": "desi",
        "cpus-per-task": 128,
    }

    default_srun_options = {
        "nodes": 1,  # N
        "ntasks": 1,  # n
        "cpus-per-task": 128,
    }


class SlurmPerlmutterTasker(SlurmTasker):
    default_header = {
        "qos": "regular",
        "nodes": 1,
        "time": "00:30:00",
        "constraint": "cpu",
        "account": "desi",
        "ntasks-per-node": 1,
    }

    default_srun_options = {}


class BashTasker(Tasker):
    """Object to write and run jobs.

    Attributes:
        command (str): Command to be run in the job.
        command_args (str): Arguments to the command.
        environment (str): Conda/python environment to load before running the command.
        environmental_variables (dict): Environmental variables to set before running the job. Format: {'environmental_variable': 'value'}.
    """

    def __init__(self, *args: Any, **kwargs: Any):
        """
        Args:
            command (str): Command to be run in the job.
            command_args (str): Arguments to the command.
            environment (str): Conda/python environment to load before running the command.
            run_file (Path): Location of the job file.
            environmental_variables (dict, optional): Environmental variables to set before running the job. Format: {'environmental_variable': 'value'}. Default: No environmental variables defined.
        """
        super().__init__(*args, **kwargs)
        if self.wait_for != None:
            raise ValueError("BachTasker won't work with wait_for feature.")

    def _make_header(self) -> str:
        """Dummy method to generate an empty header and keep compatibility with Tasker default class."""
        return ""

    def _make_env_opts(self) -> str:
        """Method to generate environmental options.

        Returns:
            str: environmental options."""
        if self.OMP_threads is not None:
            text = f"export OMP_NUM_THREADS={self.OMP_threads}\n"
        else:
            text = ""

        if "sh" in self.environment:
            activate = ""
        else:
            activate = "activate "

        text += textwrap.dedent(
            f"""
module load python
source {activate}{self.environment}
umask 0002

"""
        )

        text += self.precommand + "\n"

        return text

    def _make_run_command(self) -> str:
        """Method to genearte the run command line.

        Returns:
            str: run command line."""
        return f"$command\n"

    def send_job(self) -> None:
        """Method to run bash job script."""
        out = open(
            self.slurm_header_args["output"].replace(
                "%j", datetime.now().strftime("%d%m%Y%H%M%S")
            ),
            "w",
        )
        err = open(
            self.slurm_header_args["error"].replace(
                "%j", datetime.now().strftime("%d%m%Y%H%M%S")
            ),
            "w",
        )

        _ = Path(".").resolve()
        os.chdir(self.run_file.parent)
        self.retcode = run(["sh", f"{self.run_file}"], stdout=out, stderr=err)
        os.chdir(_)


class ChainedTasker:
    """Object to run chained Taskers.

    Attributes:
        taskers (tasker.Tasker or list of tasker.Tasker): tasker objects associated with the class
    """

    def __init__(self, taskers: List[Tasker]):
        """
        Args:
            taskers (tasker.Tasker or list of tasker.Tasker): tasker objects associated with the class
        """
        self.taskers = taskers

    def write_job(self) -> None:
        """Method to write jobs associated with taskers."""
        for tasker in self.taskers:
            tasker.write_job()

    def send_job(self) -> None:
        """Method to send jobs associated with taskers."""
        for tasker in self.taskers:
            tasker.send_job()

    @property
    def jobid(self) -> Optional[int]:
        return self.taskers[-1].jobid


class DummyTasker(Tasker):
    """Tasker object that performs no action. Useful for when files
    are copied and runs are not needed.
    """

    def __init__(self, *args: Any, **kwargs: Any):
        pass

    def write_job(self) -> None:
        pass

    def send_job(self) -> None:
        self.jobid = None
        return None



"""Script to print bookkeeper path."""

from __future__ import annotations

import argparse
from pathlib import Path
from typing import TYPE_CHECKING

from picca_bookkeeper.bookkeeper import Bookkeeper

if TYPE_CHECKING:
    from typing import Optional


def main(args: Optional[argparse.Namespace] = None) -> None:
    if args is None:
        args = getArgs()

    paths = []
    for bookkeeper_config in args.bookkeeper_configs:
        bookkeeper = Bookkeeper(bookkeeper_config)

        if bookkeeper.paths.fit_out_fname().is_file():
            paths.append(bookkeeper.paths.fit_out_fname())
        elif bookkeeper.paths.fits_path.is_dir():
            paths.append(bookkeeper.paths.fits_path)
        else:
            paths.append(bookkeeper.paths.run_path)

    if args.pretty:
        print(
            "\n\n".join(
                f"{config}:\n\t{path}"
                for config, path in zip(args.bookkeeper_configs, paths)
            )
        )
    else:
        [print(x) for x in paths]


def getArgs() -> argparse.Namespace:
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "bookkeeper_configs",
        type=Path,
        nargs="+",
        help="Bookkeeper configuration file.",
    )

    parser.add_argument(
        "--pretty", action="store_true", help="Print file name and bookkeeper paths."
    )

    args = parser.parse_args()
    return args


if __name__ == "__main__":
    main()



"""Script to fully compare two bookkeeper's configurations"""

from __future__ import annotations

import argparse
import logging
import re
import sys
from pathlib import Path
from typing import TYPE_CHECKING

from picca_bookkeeper.bookkeeper import Bookkeeper
from picca_bookkeeper.scripts.fix_bookkeeper_links import strCyan, strRed

if TYPE_CHECKING:
    from typing import Optional

logger = logging.getLogger(__name__)


def main(args: Optional[argparse.Namespace] = None) -> None:
    if args is None:
        args = getArgs()

    level = logging.getLevelName(args.log_level)
    logging.basicConfig(
        stream=sys.stdout,
        level=level,
        format="%(levelname)s:%(message)s",
    )

    bookkeeper1 = Bookkeeper(args.bookkeeper1_config)
    bookkeeper2 = Bookkeeper(args.bookkeeper2_config)

    if bookkeeper1.paths.healpix_data != bookkeeper2.paths.healpix_data:
        print(
            "Different healpix data:"
            + strRed(f"\n\t-{bookkeeper1.paths.healpix_data}")
            + strCyan(f"\n\t+{bookkeeper2.paths.healpix_data}\n")
        )

    if bookkeeper1.paths.catalog != bookkeeper2.paths.catalog:
        print(
            "Different QSO catalog:"
            + strRed(f"\n\t-{bookkeeper1.paths.catalog}")
            + strCyan(f"\n\t+{bookkeeper2.paths.catalog}\n")
        )

    if bookkeeper1.paths.catalog_dla != bookkeeper2.paths.catalog_dla:
        print(
            "Different DLA catalog:"
            + strRed(f"\n\t-{bookkeeper1.paths.catalog_dla}")
            + strCyan(f"\n\t+{bookkeeper2.paths.catalog_dla}\n")
        )

    if bookkeeper1.paths.catalog_bal != bookkeeper2.paths.catalog_bal:
        print(
            "Different BAL catalog:"
            + strRed(f"\n\t-{bookkeeper1.paths.catalog_bal}")
            + strCyan(f"\n\t+{bookkeeper2.paths.catalog_bal}\n")
        )

    tracer_catalogs_1 = (
        bookkeeper1.config.get("correlations", dict())
        .get("tracer catalogs", dict())
        .copy()
    )
    tracer_catalogs_2 = (
        bookkeeper2.config.get("correlations", dict())
        .get("tracer catalogs", dict())
        .copy()
    )

    for tracer in set(tracer_catalogs_1.keys()).union(tracer_catalogs_2.keys()):
        cat_1 = tracer_catalogs_1.get(tracer, bookkeeper1.paths.catalog)
        cat_2 = tracer_catalogs_2.get(tracer, bookkeeper2.paths.catalog)

        if cat_1 != cat_2:
            print(
                "Different tracer catalog:"
                + strRed(f"\n\t-{cat_1}")
                + strCyan(f"\n\t+{cat_2}\n")
            )

    ini_files = list(
        (bookkeeper2.paths.delta_extraction_path / "configs").glob("*.ini")
    )
    ini_files_base = [
        (bookkeeper1.paths.delta_extraction_path / "configs") / x.name
        for x in ini_files
    ]

    script_files = list((bookkeeper2.paths.correlations_path / "scripts").glob("*.sh"))
    script_files_base = [
        (bookkeeper1.paths.correlations_path / "scripts") / x.name for x in script_files
    ]

    fit_files = list((bookkeeper2.paths.fits_path / "configs").glob("*.ini"))
    fit_files_base = [
        (bookkeeper1.paths.fits_path / "configs") / x.name for x in fit_files
    ]

    config_files = ini_files + script_files + fit_files
    config_files_base = ini_files_base + script_files_base + fit_files_base

    logger.info("Number of config files: %s", len(config_files))

    import difflib

    for config_file, config_file_base in zip(config_files, config_files_base):
        if config_file_base.is_file():
            config_text = config_file_base.read_text()
        else:
            config_text = ""

        base_text = replace_strings(
            config_text,
            bookkeeper1,
        )

        config_text = replace_strings(
            config_file.read_text(),
            bookkeeper2,
        )

        # command = rf'diff <(echo "{base_text}") <(echo "{config_text}") --color --palette="ad=1;3;38;5;135:de=1;3;38;5;9"'
        # run(
        #     command,
        #     shell=True,
        #     executable='/bin/bash'
        # )

        diff_lines = list(
            difflib.unified_diff(
                base_text.split("\n"),
                config_text.split("\n"),
                fromfile="bookkeeper1",
                tofile="bookkeeper2",
                lineterm="",
                n=0,
            )
        )

        if len(diff_lines) > 0:
            print(config_file.name)
            for line in diff_lines:
                for prefix in ("---", "+++"):
                    if line.startswith(prefix):
                        break
                else:
                    if line.startswith("-"):
                        print("\t" + strRed(line))
                    elif line.startswith("+"):
                        print("\t" + strCyan(line))
                    else:
                        print("\t" + line)

            print("\n")


def replace_strings(text: str, bookkeeper: Bookkeeper) -> str:
    """Replaces strings for catalogs and bookkeeper paths so they don't
    show up in the diff
    """
    originals = [
        str(bookkeeper.paths.fits_path),
        str(bookkeeper.paths.correlations_path),
        str(bookkeeper.paths.delta_extraction_path),
        str(bookkeeper.paths.catalog),
        str(bookkeeper.paths.catalog_dla),
        str(bookkeeper.paths.catalog_bal),
        r".*ini files =.*",
        r".*zeff =.*",
        "\(",
        "\)",
        "--",
    ]

    replacements = [
        "fits-path",
        "correlations-path",
        "deltas-path",
        "catalog",
        "catalog-dla",
        "catalog-bal",
        # "catalog-tracer",
        "ini files",
        "zeff",
        "",
        "",
        "'\n'",
    ]

    for original, replacement in zip(originals, replacements):
        text = re.sub(original, replacement, text)

    for tracer, path in (
        bookkeeper.config.get("correlations", dict())
        .get("tracer catalogs", dict())
        .values()
    ):
        text = re.sub(path, tracer, text)

    return text


def getArgs() -> argparse.Namespace:
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "bookkeeper1_config",
        type=Path,
    )

    parser.add_argument(
        "bookkeeper2_config",
        type=Path,
    )

    parser.add_argument(
        "--log-level",
        default="INFO",
        choices=["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG", "NOTSET"],
    )

    return parser.parse_args()


if __name__ == "__main__":
    main()



"""Simple script to print default values"""

from __future__ import annotations

import argparse
from typing import TYPE_CHECKING

import yaml
from importlib_resources import files

from picca_bookkeeper import resources
from picca_bookkeeper.dict_utils import DictUtils

if TYPE_CHECKING:
    from typing import Optional


def main(args: Optional[argparse.Namespace] = None) -> None:
    if args is None:
        args = get_args()

    defaults = yaml.safe_load(
        files(resources).joinpath(f"default_configs/{args.prefix}.yaml").read_text()
    )

    print(DictUtils.print_dict(defaults))


def get_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument("prefix", type=str, help="Prefix to print defaults of.")
    return parser.parse_args()


if __name__ == "__main__":
    main()



""" FakeBookkeeper class to use for plotting non-bookkeeper runs"""

from __future__ import annotations

import logging
from pathlib import Path
from typing import TYPE_CHECKING

from picca_bookkeeper.bookkeeper import Bookkeeper, PathBuilder

if TYPE_CHECKING:
    from typing import Dict, Optional

logger = logging.getLogger(__name__)


class FakeBookkeeper(Bookkeeper):
    def __init__(
        self,
        attributes_files: Dict = dict(),
        export_files: Dict = dict(),
        fit_file: Optional[Path | str] = None,
    ):
        self.paths = FakePaths(
            attributes_files=attributes_files,
            export_files=export_files,
            fit_file=fit_file,
        )


class FakePaths(PathBuilder):
    def __init__(
        self,
        attributes_files: Dict = dict(),
        export_files: Dict = dict(),
        fit_file: Optional[Path | str] = None,
    ):
        self.attributes_files = attributes_files
        self.export_files = export_files
        if fit_file is not None:
            self.fit_file = Path(fit_file)

    def delta_attributes_file(
        self, region: Optional[str] = None, calib_step: Optional[int] = None
    ) -> Path:
        return Path(self.attributes_files[region])

    def exp_cf_fname(
        self,
        absorber: str,
        region: str,
        absorber2: Optional[str] = None,
        region2: Optional[str] = None,
    ) -> Path:
        if absorber is None:
            absorber = "lya"
        if region is None:
            region = "lya"
        if absorber2 is None:
            absorber2 = "lya"
        if region2 is None:
            region2 = "lya"

        return Path(self.export_files[f"{absorber}{region}_{absorber2}{region2}"])

    def exp_xcf_fname(self, absorber: str, region: str, tracer: str = "qso") -> Path:
        if absorber is None:
            absorber = "lya"
        if region is None:
            region = "lya"
        return Path(self.export_files[f"{tracer}_{absorber}{region}"])

    def fit_out_fname(self) -> Path:
        return self.fit_file



"""Simple script to cancel all jobs in a jobid log file"""

from __future__ import annotations

import argparse
from pathlib import Path
from subprocess import run
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Optional


def main(args: Optional[argparse.Namespace] = None) -> None:
    if args is None:
        args = get_args()

    with open(args.file, "r") as file:
        lines = file.readlines()

    for line in lines:
        jobid = line.split(" ")[-1].split("\n")[0]
        run(["scancel", str(jobid)])


def get_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument("file", type=Path, help="Prefix to print defaults of.")
    return parser.parse_args()


if __name__ == "__main__":
    main()



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

import matplotlib.pyplot as plt
import numpy as np
from mpl_toolkits.axes_grid1 import make_axes_locatable

from picca_bookkeeper.read_deltas import ReadDeltas

if TYPE_CHECKING:
    from typing import Dict, Optional, Tuple, Type

    from picca_bookkeeper.hints import Axes, Figure


class HeatmapAnalysis(ReadDeltas):
    pass


class Plots:
    @staticmethod
    def var_lss(
        analysis: Type[ReadDeltas],
        fig: Optional[Figure] = None,
        ax: Optional[Axes] = None,
        y_min: Optional[float] = None,
        y_max: Optional[float] = None,
        add_mean_line: bool = False,
        imshow_kwargs: Dict = dict(),
        mean_line_kwargs: Dict = dict(),
        mask: Optional[np.ndarray] = None,
        output_prefix: Optional[Path | str] = None,
        save_data: bool = False,
        save_plot: bool = False,
        save_dict: Dict = dict(),
    ) -> None:
        if (save_data or save_plot) and output_prefix is None:
            raise ValueError("Set output_prefix in order to save data.")

        if save_data:
            data_dict = {}

        if mask is None:
            if hasattr(analysis, "msk"):
                mask = analysis.msk
            else:
                mask = np.full_like(analysis.deltas2_arr, True, dtype=bool)

        n_forests = len(analysis.deltas2_arr)

        heatmap, xedges, yedges = np.histogram2d(
            np.concatenate([analysis.lambda_grid[mask[i]] for i in range(n_forests)]),
            (
                1 / analysis.weights_arr[mask]
                - (analysis.eta * analysis.var_pipe_arr)[mask]
            ),
            bins=776,
        )

        extent = (xedges[0], xedges[-1], yedges[0], yedges[-1])

        if save_data:
            data_dict["histogram-heatmap"] = heatmap
            data_dict["histogram-extent"] = extent

        if ax is None:
            fig, ax = plt.subplots()
        elif fig is None:
            raise ValueError("ax and fig should be provided at the same time")

        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)

        im = ax.imshow(
            heatmap.T,
            extent=extent,
            **{
                **dict(origin="lower", aspect="auto"),
                **imshow_kwargs,
            },
        )

        ax.set_xlabel(r"$\lambda$")
        ax.set_ylabel(r"$VAR-\eta \cdot \sigma_{\rm pip}^2$")
        ax.set_title(r"Heatmap for values of $VAR - \eta \cdot \sigma_{\rm pip}^2$")

        fig.colorbar(im, cax=cax, orientation="vertical")

        if add_mean_line:
            ycenters = 0.5 * (yedges[1:] + yedges[:-1])
            averages = []
            for i in range(len(heatmap)):
                try:
                    averages.append(np.average(ycenters, weights=heatmap[i]))
                except ZeroDivisionError:
                    averages.append(np.nan)

            ax.plot(
                0.5 * (xedges[:-1] + xedges[1:]),
                averages,
                **{
                    **dict(color="white", lw=1),
                    **mean_line_kwargs,
                },
            )

            if save_data:
                data_dict["mean_line"] = (0.5 * (xedges[:-1] + xedges[1:]), averages)

        plt.tight_layout()

        if output_prefix is not None:
            if save_plot:
                output_prefix = Path(output_prefix)
                plt.savefig(
                    output_prefix.parent / (output_prefix.name + "-var_lss.png"),
                    dpi=300,
                )

            if save_data:
                output_prefix = Path(output_prefix)
                np.savez(
                    output_prefix.parent / (output_prefix.name + "-var_lss.npz"),
                    **{**save_dict, **data_dict},
                )
        elif save_data or save_plot:
            raise ValueError("Set output_prefix in order to save data.")

    @staticmethod
    def heatmap_stat(
        analysis: Type[ReadDeltas],
        x_values: np.ndarray,
        y_stat: np.ndarray,
        bins: Tuple[int, int] | Tuple[np.ndarray, int] = (200, 777),
        use_weights: bool = False,
        fig: Optional[Figure] = None,
        ax: Optional[Axes] = None,
        y_min: Optional[float] = None,
        y_max: Optional[float] = None,
        add_mean_line: bool = False,
        imshow_kwargs: Dict = dict(),
        mean_line_kwargs: Dict = dict(),
        mask: Optional[np.ndarray] = None,
        output_prefix: Optional[Path | str] = None,
        save_data: bool = False,
        save_plot: bool = False,
        save_dict: Dict = dict(),
    ) -> None:
        """
        Arguments
        ---------
            bins (array): Which bins to use for the x-axis heatmap. (Default: (200, 777))
            use_weights (bool): Use weights instead of counts for the heatmap. (Default: False)
        """
        if (save_data or save_plot) and output_prefix is None:
            raise ValueError("Set output_prefix in order to save data.")

        if save_data:
            data_dict = {}

        if mask is None:
            if hasattr(analysis, "msk"):
                mask = analysis.msk
            else:
                mask = np.full_like(analysis.deltas2_arr, True, dtype=bool)

        if ax is None:
            fig, ax = plt.subplots()
        elif fig is None:
            raise ValueError("ax and fiug should be provided at the same time")

        n_forests = len(analysis.deltas2_arr)
        n_bins = bins[1]
        assert n_bins % 2  # This is to be able to define y_edges around mean properly

        y_center = np.average(y_stat, weights=analysis.weights_arr[mask])
        if y_min is None:
            y_min = min(
                np.percentile(y_stat, 0.1),
                y_center - np.sqrt(np.cov(y_stat, aweights=analysis.weights_arr[mask])),
            )
        if y_max is None:
            y_max = max(
                np.percentile(y_stat, 99.9),
                y_center + np.sqrt(np.cov(y_stat, aweights=analysis.weights_arr[mask])),
            )
        y_delta = (y_max - y_min) / (n_bins + 1)

        y_edges = np.concatenate(
            (
                np.flip(np.arange(y_center - y_delta / 2, y_min, -y_delta)),
                np.arange(y_center + y_delta / 2, y_max, y_delta),
            )
        )

        if use_weights:
            weights = analysis.weights_arr[mask]
        else:
            weights = None
        heatmap, xedges, yedges = np.histogram2d(  # type: ignore
            x_values,
            y_stat,
            bins=(bins[0], y_edges),
            weights=weights,
            normed=False,
        )

        extent = (xedges[0], xedges[-1], yedges[0], yedges[-1])

        if save_data:
            data_dict["histogram-heatmap"] = heatmap
            data_dict["histogram-extent"] = extent

        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)

        im = ax.imshow(
            heatmap.T,
            extent=extent,
            **{
                **dict(origin="lower", aspect="auto"),
                **imshow_kwargs,
            },
        )

        fig.colorbar(im, cax=cax, orientation="vertical")

        if add_mean_line:
            ycenters = 0.5 * (yedges[1:] + yedges[:-1])
            averages = []
            for i in range(len(heatmap)):
                try:
                    averages.append(np.average(ycenters, weights=heatmap[i]))
                except ZeroDivisionError:
                    averages.append(np.nan)

            ax.plot(
                0.5 * (xedges[:-1] + xedges[1:]),
                averages,
                **{
                    **dict(color="white", lw=1),
                    **mean_line_kwargs,
                },
            )

            if save_data:
                data_dict["mean_line"] = (0.5 * (xedges[:-1] + xedges[1:]), averages)

        if output_prefix is not None:
            if save_plot:
                output_prefix = Path(output_prefix)
                plt.tight_layout()
                plt.savefig(
                    str(output_prefix) + ".png",
                    dpi=300,
                )

            if save_data:
                np.savez(
                    str(output_prefix) + ".npz",
                    **{**save_dict, **data_dict},
                )
        elif save_data or save_plot:
            raise ValueError("Set output_prefix in order to save data.")

    @staticmethod
    def flux(
        analysis: Type[ReadDeltas],
        use_weights: bool = False,
        fig: Optional[Figure] = None,
        ax: Optional[Axes] = None,
        y_min: Optional[float] = None,
        y_max: Optional[float] = None,
        add_mean_line: bool = False,
        imshow_kwargs: Dict = dict(),
        mean_line_kwargs: Dict = dict(),
        mask: Optional[np.ndarray] = None,
        output_prefix: Optional[Path | str] = None,
        save_data: bool = False,
        save_plot: bool = False,
        save_dict: Dict = dict(),
    ) -> None:
        if output_prefix is not None:
            output_prefix = Path(output_prefix).parent / (
                Path(output_prefix).name + "-flux"
            )

        if mask is None:
            if hasattr(analysis, "msk"):
                mask = analysis.msk
            else:
                mask = np.full_like(analysis.deltas2_arr, True, dtype=bool)

        if ax is None:
            fig, ax = plt.subplots()
        elif fig is None:
            raise ValueError("ax and fig should be provided at the same time")

        Plots.heatmap_stat(
            analysis,
            x_values=np.concatenate(
                [
                    analysis.lambda_grid[mask[i]]
                    for i in range(len(analysis.deltas2_arr))
                ]
            ),
            y_stat=np.log10(analysis.flux_arr[mask]),
            use_weights=use_weights,
            bins=(analysis.lambda_edges, 777),
            fig=fig,
            ax=ax,
            y_min=y_min,
            y_max=y_max,
            add_mean_line=add_mean_line,
            imshow_kwargs=imshow_kwargs,
            mean_line_kwargs=mean_line_kwargs,
            mask=mask if mask is not None else None,
            output_prefix=output_prefix,
            save_data=save_data,
            save_dict=save_dict,
        )
        ax.set_xlabel(r"$\lambda$")
        ax.set_ylabel(r"Flux")
        ax.set_title(r"Heatmap for values of flux")

        if save_plot:
            plt.tight_layout()
            plt.savefig(
                str(output_prefix) + ".png",
                dpi=300,
            )

    @staticmethod
    def flux_var(
        analysis: Type[ReadDeltas],
        use_weights: bool = False,
        fig: Optional[Figure] = None,
        ax: Optional[Axes] = None,
        y_min: Optional[float] = None,
        y_max: Optional[float] = None,
        add_mean_line: bool = False,
        imshow_kwargs: Dict = dict(),
        mean_line_kwargs: Dict = dict(),
        mask: Optional[np.ndarray] = None,
        output_prefix: Optional[Path | str] = None,
        save_data: bool = False,
        save_plot: bool = False,
        save_dict: Dict = dict(),
    ) -> None:
        if output_prefix is not None:
            output_prefix = Path(output_prefix).parent / (
                Path(output_prefix).name + "-flux_var"
            )

        if mask is None:
            if hasattr(analysis, "msk"):
                mask = analysis.msk
            else:
                mask = np.full_like(analysis.deltas2_arr, True, dtype=bool)

        if ax is None:
            fig, ax = plt.subplots()
        elif fig is None:
            raise ValueError("ax and fig should be provided at the same time")

        Plots.heatmap_stat(
            analysis,
            x_values=np.concatenate(
                [
                    analysis.lambda_grid[mask[i]]
                    for i in range(len(analysis.deltas2_arr))
                ]
            ),
            y_stat=np.log10(1 / (analysis.flux_ivar_arr[mask])),
            use_weights=use_weights,
            bins=(analysis.lambda_edges, 777),
            fig=fig,
            ax=ax,
            y_min=y_min,
            y_max=y_max,
            add_mean_line=add_mean_line,
            imshow_kwargs=imshow_kwargs,
            mean_line_kwargs=mean_line_kwargs,
            output_prefix=output_prefix,
            save_data=save_data,
            save_dict=save_dict,
        )
        ax.set_xlabel(r"$\lambda$")
        ax.set_ylabel(r"$\log(\sigma_{\rm pip})$")
        ax.set_title(r"Heatmap for values of $\sigma_{\rm pip}$")

        if save_plot:
            plt.tight_layout()
            plt.savefig(
                str(output_prefix) + ".png",
                dpi=300,
            )

    @staticmethod
    def flux_var_over_flux(
        analysis: Type[ReadDeltas],
        use_weights: bool = False,
        fig: Optional[Figure] = None,
        ax: Optional[Axes] = None,
        y_min: Optional[float] = None,
        y_max: Optional[float] = None,
        add_mean_line: bool = False,
        imshow_kwargs: Dict = dict(),
        mean_line_kwargs: Dict = dict(),
        mask: Optional[np.ndarray] = None,
        output_prefix: Optional[Path | str] = None,
        save_data: bool = False,
        save_plot: bool = False,
        save_dict: Dict = dict(),
    ) -> None:
        if output_prefix is not None:
            output_prefix = Path(output_prefix).parent / (
                Path(output_prefix).name + "-flux_var"
            )

        if mask is None:
            if hasattr(analysis, "msk"):
                mask = analysis.msk
            else:
                mask = np.full_like(analysis.deltas2_arr, True, dtype=bool)

        if ax is None:
            fig, ax = plt.subplots()
        elif fig is None:
            raise ValueError("ax and fig should be provided at the same time")

        y_stat = np.log10(1 / (analysis.flux_ivar_arr[mask] * analysis.flux_arr[mask]))
        # y_stat=1/(analysis.flux_ivar_arr[mask]*analysis.flux_arr[mask])
        Plots.heatmap_stat(
            analysis,
            x_values=np.concatenate(
                [
                    analysis.lambda_grid[mask[i]]
                    for i in range(len(analysis.deltas2_arr))
                ]
            ),
            y_stat=y_stat,  # np.log10(1 / analysis.flux_ivar_arr[mask]),
            use_weights=use_weights,
            bins=(analysis.lambda_edges, 777),
            fig=fig,
            ax=ax,
            y_min=y_min,
            y_max=y_max,
            add_mean_line=add_mean_line,
            imshow_kwargs=imshow_kwargs,
            mean_line_kwargs=mean_line_kwargs,
            output_prefix=output_prefix,
            save_data=save_data,
            save_dict=save_dict,
        )
        ax.set_xlabel(r"$\lambda$")
        ax.set_ylabel(r"$\log(\sigma_{\rm pip}/Flux)")
        ax.set_title(r"Heatmap for values of $\sigma_{\rm pip}$/Flux")

        if save_plot:
            plt.tight_layout()
            plt.savefig(
                str(output_prefix) + ".png",
                dpi=300,
            )

    @staticmethod
    def pipe_var(
        analysis: Type[ReadDeltas],
        use_weights: bool = False,
        fig: Optional[Figure] = None,
        ax: Optional[Axes] = None,
        y_min: Optional[float] = None,
        y_max: Optional[float] = None,
        add_mean_line: bool = False,
        imshow_kwargs: Dict = dict(),
        mean_line_kwargs: Dict = dict(),
        mask: Optional[np.ndarray] = None,
        output_prefix: Optional[Path | str] = None,
        save_data: bool = False,
        save_plot: bool = False,
        save_dict: Dict = dict(),
    ) -> None:
        if output_prefix is not None:
            output_prefix = Path(output_prefix).parent / (
                Path(output_prefix).name + "-pipe_var"
            )

        if mask is None:
            if hasattr(analysis, "msk"):
                mask = analysis.msk
            else:
                mask = np.full_like(analysis.deltas2_arr, True, dtype=bool)

        if ax is None:
            fig, ax = plt.subplots()
        elif fig is None:
            raise ValueError("ax and fig should be provided at the same time")

        Plots.heatmap_stat(
            analysis,
            x_values=np.concatenate(
                [
                    analysis.lambda_grid[mask[i]]
                    for i in range(len(analysis.deltas2_arr))
                ]
            ),
            y_stat=np.log10(analysis.var_pipe_arr[mask]),
            use_weights=use_weights,
            bins=(analysis.lambda_edges, 777),
            fig=fig,
            ax=ax,
            y_min=y_min,
            y_max=y_max,
            add_mean_line=add_mean_line,
            imshow_kwargs=imshow_kwargs,
            mean_line_kwargs=mean_line_kwargs,
            output_prefix=output_prefix,
            save_data=save_data,
            save_dict=save_dict,
        )
        ax.set_xlabel(r"$\lambda$")
        ax.set_ylabel(r"$\log(\tilde{\sigma}_{\rm pip})$")
        ax.set_title(r"Heatmap for values of $\tilde{\sigma}_{\rm pip}$")

        if save_plot:
            plt.tight_layout()
            plt.savefig(
                str(output_prefix) + ".png",
                dpi=300,
            )

    @staticmethod
    def flux_var_rf(
        analysis: Type[ReadDeltas],
        use_weights: bool = False,
        fig: Optional[Figure] = None,
        ax: Optional[Axes] = None,
        y_min: Optional[float] = None,
        y_max: Optional[float] = None,
        add_mean_line: bool = False,
        imshow_kwargs: Dict = dict(),
        mean_line_kwargs: Dict = dict(),
        mask: Optional[np.ndarray] = None,
        output_prefix: Optional[Path | str] = None,
        save_data: bool = False,
        save_plot: bool = False,
        save_dict: Dict = dict(),
    ) -> None:
        if output_prefix is not None:
            output_prefix = Path(output_prefix).parent / (
                Path(output_prefix).name + "-flux-_var_rf"
            )

        if mask is None:
            if hasattr(analysis, "msk"):
                mask = analysis.msk
            else:
                mask = np.full_like(analysis.deltas2_arr, True, dtype=bool)

        if ax is None:
            fig, ax = plt.subplots()
        elif fig is None:
            raise ValueError("ax and fig should be provided at the same time")

        Plots.heatmap_stat(
            analysis,
            x_values=analysis.lambda_rf_arr[mask],
            y_stat=np.log10(1 / analysis.flux_ivar_arr[mask]),
            use_weights=use_weights,
            bins=(200, 777),
            fig=fig,
            ax=ax,
            y_min=y_min,
            y_max=y_max,
            add_mean_line=add_mean_line,
            imshow_kwargs=imshow_kwargs,
            mean_line_kwargs=mean_line_kwargs,
            output_prefix=output_prefix,
            save_data=save_data,
            save_dict=save_dict,
        )

        ax.set_xlabel(r"$\lambda$")
        ax.set_ylabel(r"$\log(\sigma_{\rm pip})$")
        ax.set_title(r"Heatmap for values of $\sigma_{\rm pip}$")

        if save_plot:
            plt.tight_layout()
            plt.savefig(
                str(output_prefix) + ".png",
                dpi=300,
            )

    @staticmethod
    def lambda_vs_lambda_rf(
        analysis: Type[ReadDeltas],
        use_weights: bool = False,
        fig: Optional[Figure] = None,
        ax: Optional[Axes] = None,
        y_min: Optional[float] = None,
        y_max: Optional[float] = None,
        add_mean_line: bool = False,
        imshow_kwargs: Dict = dict(),
        mean_line_kwargs: Dict = dict(),
        mask: Optional[np.ndarray] = None,
        output_prefix: Optional[Path | str] = None,
        save_data: bool = False,
        save_plot: bool = False,
        save_dict: Dict = dict(),
    ) -> None:
        if output_prefix is not None:
            output_prefix = Path(output_prefix).parent / (
                Path(output_prefix).name + "-lambda_vs_lambda_rf"
            )

        if mask is None:
            if hasattr(analysis, "msk"):
                mask = analysis.msk
            else:
                mask = np.full_like(analysis.deltas2_arr, True, dtype=bool)

        if ax is None:
            fig, ax = plt.subplots()
        elif fig is None:
            raise ValueError("ax and fig should be provided at the same time")

        Plots.heatmap_stat(
            analysis,
            x_values=np.concatenate(
                [
                    analysis.lambda_grid[mask[i]]
                    for i in range(len(analysis.deltas2_arr))
                ]
            ),
            y_stat=analysis.lambda_rf_arr[mask],
            use_weights=use_weights,
            bins=(analysis.lambda_edges, 777),
            fig=fig,
            ax=ax,
            y_min=y_min,
            y_max=y_max,
            add_mean_line=add_mean_line,
            imshow_kwargs=imshow_kwargs,
            mean_line_kwargs=mean_line_kwargs,
            output_prefix=output_prefix,
            save_data=save_data,
            save_dict=save_dict,
        )

        ax.set_xlabel(r"$\lambda$")
        ax.set_ylabel(r"$\lambda_{\rm RF}$")

        if save_plot:
            plt.tight_layout()
            plt.savefig(
                str(output_prefix) + ".png",
                dpi=300,
            )

    @staticmethod
    def var_residual(
        analysis: Type[ReadDeltas],
        fig: Optional[Figure] = None,
        ax: Optional[Axes] = None,
    ) -> None:
        pass



""" Script to run picca_xcf and export given a bookkeeper config file."""

from __future__ import annotations

import argparse
import logging
import sys
from pathlib import Path
from typing import TYPE_CHECKING

from picca_bookkeeper.bookkeeper import Bookkeeper
from picca_bookkeeper.tasker import DummyTasker

if TYPE_CHECKING:
    from typing import Optional

logger = logging.getLogger(__name__)


def main(args: Optional[argparse.Namespace] = None) -> None:
    if args is None:
        args = get_args()

    level = logging.getLevelName(args.log_level)
    logging.basicConfig(
        stream=sys.stdout,
        level=level,
        format="%(levelname)s:%(message)s",
    )

    logger.info(f"Adding cross-correlation: {args.absorber}{args.region}_{args.tracer}")

    bookkeeper = Bookkeeper(
        args.bookkeeper_config,
        overwrite_config=args.overwrite_config,
        read_mode=False,
    )

    xcf = bookkeeper.get_xcf_tasker(
        region=args.region,
        absorber=args.absorber,
        tracer=args.tracer,
        debug=args.debug,
        system=args.system,
        wait_for=args.wait_for,
        overwrite=args.overwrite,
        skip_sent=args.skip_sent,
    )
    xcf.write_job()
    if not args.only_write:
        xcf.send_job()

        if not isinstance(xcf, DummyTasker):
            logger.info(
                "Sent cross-correlation "
                f"{args.absorber}{args.region}_{args.tracer}:\n\t"
                f"{xcf.jobid}"
            )
    logger.info("Done.\n")

    if not bookkeeper.config.get("fits", dict()).get("no distortion", False):
        logger.info(
            f"Adding distortion matrix: {args.absorber}{args.region}_{args.tracer}"
        )
        xdmat = bookkeeper.get_xdmat_tasker(
            region=args.region,
            absorber=args.absorber,
            tracer=args.tracer,
            debug=args.debug,
            system=args.system,
            wait_for=args.wait_for,
            overwrite=args.overwrite,
            skip_sent=args.skip_sent,
        )
        xdmat.write_job()
        if not args.only_write:
            xdmat.send_job()

            if not isinstance(xdmat, DummyTasker):
                logger.info(
                    "Sent distortion matrix "
                    f"{args.absorber}{args.region}_{args.tracer}:\n\t"
                    f"{xdmat.jobid}"
                )
        logger.info("Done.\n")

    if not bookkeeper.config.get("fits", dict()).get(
        "no metals", False
    ) and not bookkeeper.config.get("fits", dict()).get("vega metals", False):
        # Compute metals if metals should be included and metals are not going
        # to be computed by vega.

        logger.info(f"Adding metal matrix: {args.absorber}{args.region}_{args.tracer}")
        metal = bookkeeper.get_xmetal_tasker(
            region=args.region,
            absorber=args.absorber,
            tracer=args.tracer,
            system=args.system,
            debug=args.debug,
            wait_for=args.wait_for,
            overwrite=args.overwrite,
            skip_sent=args.skip_sent,
        )
        metal.write_job()
        if not args.only_write:
            metal.send_job()

            if not isinstance(metal, DummyTasker):
                logger.info(
                    "Sent metal matrix "
                    f"{args.absorber}{args.region}_{args.tracer}:\n\t"
                    f"{metal.jobid}"
                )
        logger.info("Done.\n")

    logger.info(f"Adding export: {args.absorber}{args.region}_{args.tracer}")
    xcf_exp = bookkeeper.get_xcf_exp_tasker(
        region=args.region,
        absorber=args.absorber,
        tracer=args.tracer,
        system=args.system,
        wait_for=args.wait_for,
        overwrite=args.overwrite,
        skip_sent=args.skip_sent,
    )

    xcf_exp.write_job()
    if not args.only_write:
        xcf_exp.send_job()

        if not isinstance(xcf_exp, DummyTasker):
            logger.info(
                "Sent export "
                f"{args.absorber}{args.region}_{args.tracer}:\n\t"
                f"{xcf_exp.jobid}"
            )
    logger.info("Done.\n")


def get_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "bookkeeper_config", type=Path, help="Path to bookkeeper file to use"
    )

    parser.add_argument(
        "--system",
        type=str,
        default=None,
    )

    parser.add_argument(
        "--region",
        type=str,
        choices=["lya", "lyb"],
        default="lya",
        help="Region to compute correlation in",
    )

    parser.add_argument(
        "--tracer", type=str, default="qso", help="Tracer to use for the correlation."
    )

    parser.add_argument(
        "--absorber", type=str, default="lya", help="Absorber to use for correlations"
    )

    parser.add_argument(
        "--overwrite", action="store_true", help="Force overwrite output data."
    )

    parser.add_argument(
        "--overwrite-config",
        action="store_true",
        help="Force overwrite bookkeeper config.",
    )

    parser.add_argument(
        "--skip-sent", action="store_true", help="Skip runs that were already sent."
    )

    parser.add_argument(
        "--debug",
        action="store_true",
    )

    parser.add_argument(
        "--only-write", action="store_true", help="Only write scripts, not send them."
    )

    parser.add_argument("--wait-for", nargs="+", type=int, default=None, required=False)

    parser.add_argument(
        "--log-level",
        default="INFO",
        choices=["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG", "NOTSET"],
    )

    args = parser.parse_args()

    return args


if __name__ == "__main__":
    main()



"""
    Add column LAST_NIGHT with a default time value to a given catalog.
"""

from __future__ import annotations

import argparse
import logging
import sys
from pathlib import Path
from typing import TYPE_CHECKING

import fitsio
import numpy as np

if TYPE_CHECKING:
    from typing import Optional
logger = logging.getLogger(__name__)


def main(args: Optional[argparse.Namespace] = None) -> None:
    if args is None:
        args = getArgs()

    level = logging.getLevelName(args.log_level)
    logging.basicConfig(
        stream=sys.stdout,
        level=level,
        format="%(levelname)s:%(message)s",
    )

    logger.info(f"Using fill value: {args.fill_value}")

    with fitsio.FITS(args.input_cat, "rw") as hdul:
        data = np.asarray(
            [args.fill_value for i in range(hdul["ZCATALOG"].read_header()["NAXIS2"])]
        )
        hdul["ZCATALOG"].insert_column(name="LAST_NIGHT", data=data)

    logger.info("Done.")


def getArgs() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "input_cat",
        type=Path,
        help="Input catalog",
    )

    parser.add_argument(
        "--fill-value",
        default="20210403",
        type=int,
        help="Date to use as fill value. Format: YYYYMMDD",
    )

    parser.add_argument(
        "--log-level",
        default="INFO",
        choices=["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG", "NOTSET"],
    )

    args = parser.parse_args()
    return args


if __name__ == "__main__":
    main()



"""Script to run multiple fits one after another, ideal for interactive nodes and multiple fits (and only fits) to be run"""

from __future__ import annotations

import argparse
import logging
import sys
from pathlib import Path
from subprocess import run
from typing import TYPE_CHECKING

from picca_bookkeeper.bookkeeper import Bookkeeper
from picca_bookkeeper.dict_utils import DictUtils

if TYPE_CHECKING:
    from typing import Optional
logger = logging.getLogger(__name__)


def main(args: Optional[argparse.Namespace] = None) -> None:
    if args is None:
        args = get_args()

    level = logging.getLevelName(args.log_level)
    logging.basicConfig(
        stream=sys.stdout,
        level=level,
        format="%(levelname)s:%(message)s",
    )

    for bookkeeper_config in args.bookkeeper_configs:
        bookkeeper = Bookkeeper(
            bookkeeper_config,
            read_mode=False,
            overwrite_config=args.overwrite_config,
        )

        config = DictUtils.merge_dicts(
            bookkeeper.defaults,
            bookkeeper.config,
        )

        logger.info(f"Adding fit:{bookkeeper.paths.fits_path}")

        if config["fits"].get("compute zeff", False) and not args.skip_zeff:
            computed_params_file = bookkeeper.paths.fit_computed_params_out()
            if (
                not computed_params_file.is_file()
                or computed_params_file.stat().st_size < 20
                or args.overwrite
            ):
                # In this case we need to compute first the params
                compute_zeff = bookkeeper.get_compute_zeff_tasker(
                    overwrite=True,
                    skip_sent=False,
                )
                commands = compute_zeff._make_command().split('"')[1].strip().split(" ")
                with open(compute_zeff.slurm_header_args["output"], "w") as out, open(
                    compute_zeff.slurm_header_args["error"], "w"
                ) as err:
                    retcode = run(commands, stdout=out, stderr=err)
                logger.info(f"\tcompute zeff: {retcode.returncode}")

        fit = bookkeeper.get_fit_tasker(
            overwrite=True,
            skip_sent=False,
        )
        if bookkeeper.paths.fit_out_fname().stat().st_size < 20 or args.overwrite:
            commands = fit._make_command().split('"')[1].strip().split(" ")
            with open(fit.slurm_header_args["output"], "w") as out, open(
                fit.slurm_header_args["error"], "w"
            ) as err:
                retcode = run(commands, stdout=out, stderr=err)
            logger.info(f"\tfit: {retcode.returncode}")
        else:
            logger.info(f"\tfit already finished.")


def get_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "bookkeeper_configs",
        type=Path,
        nargs="*",
        help="Path to bookkeeper file to use",
    )

    parser.add_argument(
        "--system",
        type=str,
        default=None,
    )

    parser.add_argument(
        "--overwrite", action="store_true", help="Force overwrite output data."
    )

    parser.add_argument(
        "--overwrite-config",
        action="store_true",
        help="Force overwrite bookkeeper config.",
    )

    parser.add_argument(
        "--skip-zeff",
        action="store_true",
        help="Skip computation of zeff and related quantities.",
    )

    parser.add_argument(
        "--log-level",
        default="INFO",
        choices=["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG", "NOTSET"],
    )

    args = parser.parse_args()

    return args


if __name__ == "__main__":
    main()



"""Script to run all the analyis from terminal"""

from __future__ import annotations

import argparse
import logging
import sys
from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np

from picca_bookkeeper.bookkeeper import Bookkeeper
from picca_bookkeeper.dict_utils import DictUtils
from picca_bookkeeper.scripts.run_cf import main as run_cf
from picca_bookkeeper.scripts.run_delta_extraction import main as run_delta_extraction
from picca_bookkeeper.scripts.run_fit import main as run_fit
from picca_bookkeeper.scripts.run_sampler import main as run_sampler
from picca_bookkeeper.scripts.run_xcf import main as run_xcf

if TYPE_CHECKING:
    from typing import Optional
logger = logging.getLogger(__name__)


def main(args: Optional[argparse.Namespace] = None) -> None:
    if args is None:
        args = get_args()

    level = logging.getLevelName(args.log_level)
    logging.basicConfig(
        stream=sys.stdout,
        level=level,
        format="%(levelname)s:%(message)s",
    )

    bookkeeper = Bookkeeper(
        args.bookkeeper_config,
        overwrite_config=args.overwrite_config,
        read_mode=False,
    )

    config = DictUtils.merge_dicts(
        bookkeeper.defaults,
        bookkeeper.config,
    )

    raw = config["general"].get("raw mocks", False)
    true = config["general"].get("true mocks", False)

    ########################################
    ## Identifying needed runs from names
    ########################################
    regions = []

    autos = []
    auto_correlations = args.auto_correlations
    if config.get("fits", dict()).get("auto correlations", None) not in (None, ""):
        auto_correlations += config["fits"]["auto correlations"].split(" ")
    for auto in auto_correlations:
        absorber, region, absorber2, region2 = auto.replace("-", ".").split(".")

        region = bookkeeper.validate_region(region)
        absorber = bookkeeper.validate_absorber(absorber)
        region2 = bookkeeper.validate_region(region2)
        absorber2 = bookkeeper.validate_absorber(absorber2)

        autos.append([absorber, region, absorber2, region2])
        regions.append(region)
        regions.append(region2)

    crosses = []
    cross_correlations = args.cross_correlations
    if config.get("fits", dict()).get("cross correlations", None) not in (None, ""):
        cross_correlations += (
            config["fits"]["cross correlations"].replace("-", ".").split(" ")
        )
    for cross in cross_correlations:
        tracer, absorber, region = cross.split(".")
        region = bookkeeper.validate_region(region)
        absorber = bookkeeper.validate_absorber(absorber)

        regions.append(region)
        crosses.append([absorber, region, tracer])

    regions = np.unique(regions)

    ########################################
    ## Running delta extraction for calibration
    ## and then all the deltas needed.
    ########################################
    if not args.no_deltas:
        if (not raw and not true) and config["delta extraction"]["calib"] != 0:
            calib_args = argparse.Namespace(
                bookkeeper_config=args.bookkeeper_config,
                region="lya",  # It doesn't really matter
                overwrite_config=False,
                system=args.system,
                debug=args.debug,
                only_calibration=True,  # Because we are only running calib
                skip_calibration=False,
                only_write=args.only_write,
                wait_for=args.wait_for,
                log_level=args.log_level,
                overwrite=args.overwrite,
                skip_sent=args.skip_sent,
            )
            run_delta_extraction(calib_args)

        for region in regions:
            region_args = argparse.Namespace(
                bookkeeper_config=args.bookkeeper_config,
                region=region,
                overwrite_config=False,
                debug=args.debug,
                system=args.system,
                only_calibration=False,
                skip_calibration=True,
                only_write=args.only_write,
                wait_for=args.wait_for,
                log_level=args.log_level,
                overwrite=args.overwrite,
                skip_sent=args.skip_sent,
            )
            run_delta_extraction(region_args)

    ########################################
    ## Running all the correlations needed
    ########################################
    if not args.no_correlations:
        for auto in autos:
            absorber, region, absorber2, region2 = auto

            auto_args = argparse.Namespace(
                bookkeeper_config=args.bookkeeper_config,
                overwrite_config=False,
                region=region,
                region2=region2,
                absorber=absorber,
                absorber2=absorber2,
                system=args.system,
                debug=False,  # Debug, only set deltas
                only_write=args.only_write,
                wait_for=args.wait_for,
                log_level=args.log_level,
                overwrite=args.overwrite,
                skip_sent=args.skip_sent,
            )
            run_cf(auto_args)

        for cross in crosses:
            absorber, region, tracer = cross

            cross_args = argparse.Namespace(
                bookkeeper_config=args.bookkeeper_config,
                overwrite_config=False,
                region=region,
                absorber=absorber,
                tracer=tracer,
                system=args.system,
                debug=False,  # Debug, only set deltas,
                only_write=args.only_write,
                wait_for=args.wait_for,
                log_level=args.log_level,
                overwrite=args.overwrite,
                skip_sent=args.skip_sent,
            )
            run_xcf(cross_args)

    ########################################
    ## Running fits
    ########################################
    if not args.no_fits:
        fit_args = argparse.Namespace(
            bookkeeper_config=args.bookkeeper_config,
            overwrite_config=False,
            system=args.system,
            only_write=args.only_write,
            wait_for=args.wait_for,
            log_level=args.log_level,
            overwrite=args.overwrite,
            skip_sent=args.skip_sent,
        )
        run_fit(fit_args)

    if args.sampler:
        sampler_args = argparse.Namespace(
            bookkeeper_config=args.bookkeeper_config,
            system=args.system,
            overwrite_config=False,
            only_write=args.only_write,
            wait_for=args.wait_for,
            log_level=args.log_level,
            overwrite=args.overwrite,
            skip_sent=args.skip_sent,
        )
        run_sampler(sampler_args)


def get_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "bookkeeper_config", type=Path, help="Path to bookkeeper file to use"
    )

    parser.add_argument(
        "--system",
        type=str,
        default=None,
    )

    parser.add_argument(
        "--overwrite", action="store_true", help="Force overwrite output data."
    )

    parser.add_argument(
        "--overwrite-config",
        action="store_true",
        help="Force overwrite bookkeeper config.",
    )

    parser.add_argument(
        "--skip-sent", action="store_true", help="Skip runs that were already sent."
    )

    parser.add_argument(
        "--auto-correlations",
        type=str,
        nargs="+",
        default=[],
        help="List of auto-correlations to include in the vega "
        "fits. The format of the strings should be 'lya.lya-lya.lyb'. "
        "which reads as Lyman-alpha absorption in the Lyman-alpha region "
        "correlated with lyman alpha in the lyman beta region. "
        "This is to allow splitting.",
    )

    parser.add_argument(
        "--cross-correlations",
        type=str,
        nargs="+",
        default=[],
        help="List of cross-correlations to include in the vega "
        "fits. The format of the strings should be 'lya.lya'.",
    )

    parser.add_argument(
        "--no-deltas", action="store_true", help="Don't measure deltas."
    )

    parser.add_argument(
        "--no-correlations", action="store_true", help="Don't measure correlations."
    )

    parser.add_argument("--no-fits", action="store_true", help="Don't measure fits.")

    parser.add_argument("--sampler", action="store_true", help="Run the sampler.")

    parser.add_argument(
        "--debug",
        action="store_true",
    )

    parser.add_argument(
        "--only-write", action="store_true", help="Only write scripts, not send them."
    )

    parser.add_argument("--wait-for", nargs="+", type=int, default=None, required=False)

    parser.add_argument(
        "--log-level",
        default="INFO",
        choices=["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG", "NOTSET"],
    )

    args = parser.parse_args()

    return args


if __name__ == "__main__":
    main()



"""
    Read flux, ivar, mask and wavelength information for DESI data
"""

from __future__ import annotations

import argparse
import logging
import sys
from multiprocessing import Pool
from pathlib import Path
from typing import TYPE_CHECKING

import fitsio
import healpy as hp
import matplotlib.pyplot as plt
import numpy as np
from astropy.table import Table

if TYPE_CHECKING:
    from typing import Dict, List, Optional

logger = logging.getLogger(__name__)


def read_file(file: Path | str, valid_targetids: List[str | int]) -> np.ndarray:
    ffile = fitsio.FITS(file)
    targetid = ffile["FIBERMAP"]["TARGETID"].read()

    in1d = np.in1d(targetid, valid_targetids)

    values = []
    for color in "B", "R", "Z":
        values.append(ffile[f"{color}_WAVELENGTH"].read())
        values.append(ffile[f"{color}_FLUX"].read()[in1d])
        values.append(ffile[f"{color}_IVAR"].read()[in1d])
        values.append(ffile[f"{color}_MASK"].read()[in1d])

    values.append(targetid[in1d])

    return np.asarray(values)


class Plots:
    @staticmethod
    def masked_pixels_percentage(
        properties: Dict,
        output_prefix: Path,
        plot_kwargs: Dict = dict(),
        downsampling: float = 1,
        save_data: bool = False,
        save_plot: bool = False,
        save_dict: Dict = dict(),
    ) -> None:
        fig, axs = plt.subplots(1, 3, sharey=True, figsize=(10, 5))

        for ax, color in zip(axs, ("B", "R", "Z")):
            ax.plot(
                properties[f"{color}_WAVELENGTH"],
                (properties[f"{color}_MASK"] != 0).sum(axis=0)
                / len(properties[f"{color}_MASK"])
                * 100,
                **plot_kwargs,
            )
            ax.set_title(f"{color}-band")

        axs[0].set_ylabel(r"% of masked pixels")
        fig.supxlabel(r"$\lambda \, [\AA]$")
        if downsampling < 1:
            plt.suptitle(
                f"Percentage of masked pixels for a subsample of fugu data (subsample {round(downsampling*100, 2)}%)"
            )
        else:
            plt.suptitle("Percentage of masked pixels")

        plt.tight_layout()
        if save_plot:
            plt.savefig(
                output_prefix.parent
                / (output_prefix.name + "-masked_pixels_percentage.png"),
                dpi=300,
            )

        if save_data:
            data_dict = {}
            for color in "B", "R", "Z":
                data_dict[f"{color}_WAVELENGTH"] = properties[f"{color}_WAVELENGTH"]
                data_dict[f"{color}_VALUE"] = (
                    (properties[f"{color}_MASK"] != 0).sum(axis=0)
                    / len(properties[f"{color}_MASK"])
                    * 100
                )

            np.savez(
                output_prefix.parent
                / (output_prefix.name + "-masked_pixels_percentage.npz"),
                **{**save_dict, **data_dict},
            )

    @staticmethod
    def histogram_qsos_per_masked_bins(
        properties: Dict,
        output_prefix: Path,
        hist_kwargs: Dict = dict(),
        downsampling: float = 1,
        save_data: bool = False,
        save_plot: bool = False,
        save_dict: Dict = dict(),
    ) -> None:
        if save_data:
            data_dict = {}

        hist_kwargs = {
            **dict(bins=np.linspace(0, 80, 10)),
            **hist_kwargs,
        }
        fig, axs = plt.subplots(1, 3, sharey=True, figsize=(10, 5))

        for ax, color in zip(axs, ("B", "R", "Z")):
            hist_ret = ax.hist(
                (properties[f"{color}_MASK"] != 0).sum(axis=1),
                **hist_kwargs,
            )

            if save_data:
                data_dict[f"{color}_hist"] = hist_ret[:2]

            ax.set_title(f"{color}-band")

        fig.supxlabel("# masked pixels per QSO")
        if downsampling < 1:
            fig.suptitle(
                f"Histogram: Histogram of QSOs per number of pixel masked (subsample {round(downsampling*100, 2)}%)"
            )
        else:
            fig.suptitle("Histogram: Histogram of QSOs per number of pixel masked")

        axs[0].set_ylabel("# QSOs")
        plt.tight_layout()

        if save_plot:
            plt.savefig(
                output_prefix.parent
                / (output_prefix.name + "-qso_histogram_per_masked_bin.png"),
                dpi=300,
            )

        if save_data:
            np.savez(
                output_prefix.parent
                / (output_prefix.name + "-qso_hsitogram_per_masked_bin.npz"),
                **{**save_dict, **data_dict},
            )


class ReadDESILyaData:
    def __init__(self, input_directory: Path, catalogue: Path | str):
        """
        Read flux, ivar, mask and wavelength information for DESI data

        Arguments:
        ----------
            input_directory (Path): Input directory.
            catalogue (Path): QSO catalog.
        """
        self.input_directory = Path(input_directory)
        if not input_directory.is_dir():
            raise FileNotFoundError(f"Invalid input directory: {self.input_directory}")
        if not Path(catalogue).is_file():
            raise FileNotFoundError(f"Invalid catalogue file: {catalogue}")

        self.catalogue = Table(
            fitsio.read(
                catalogue,
                ext="ZCATALOG",
            )
        )

    def read_data(
        self,
        downsampling: float = 1,
        file_downsampling: float = 1,
        processes: Optional[int] = None,
    ) -> Dict:
        """
        Read data from DESI files.

        Arguments:
        ----------
            downsampling (float, optional): Downsampling to apply to QSO catalogue (Default: 1).
            file_downsampling (float, optional): Downsampling to apply to the number of files to read (Default: 1).
            processes (int, optional): Number of processes to use for reading files. (Default: None).
        """
        if downsampling < 1:
            logger.info(f"Downsampling QSO catalogue. p={downsampling}")
            np.random.seed(1)
            msk = np.random.choice(
                [True, False],
                len(self.catalogue),
                p=[downsampling, 1 - downsampling],
            )

            catalogue = self.catalogue[msk]
        else:
            catalogue = self.catalogue

        healpix = [
            hp.ang2pix(
                64,
                row["TARGET_RA"],
                row["TARGET_DEC"],
                lonlat=True,
                nest=True,
            )
            for row in catalogue
        ]

        catalogue["HEALPIX"] = healpix
        logger.info("Sorting catalog by HEALPIX")
        catalogue.sort("HEALPIX")

        logger.info("Grouping catalogue by HEALPIX, SURVEY")
        grouped_catalogue = catalogue.group_by(["HEALPIX", "SURVEY"])

        logger.info("Retreiving filenames")

        files_to_read = []
        group_valid_targets = []
        for group in grouped_catalogue.groups:
            healpix, survey = group["HEALPIX", "SURVEY"][0]

            filename = (
                self.input_directory
                / survey
                / "dark"
                / str(healpix // 100)  # type: ignore
                / str(healpix)  # type: ignore
                / f"coadd-{survey}-dark-{healpix}.fits"
            )

            if filename.is_file():
                files_to_read.append(filename)
                group_valid_targets.append(group["TARGETID"])

        logger.info(f"Number of files to read: {len(files_to_read)}.")
        if file_downsampling < 1:
            logger.info(f"Downsampling files to read.")
            np.random.seed(2)
            msk = np.random.choice(
                [True, False],
                len(files_to_read),
                p=[file_downsampling, 1 - file_downsampling],
            )

            files_to_read = np.asarray(files_to_read)[msk]
            group_valid_targets = np.asarray(group_valid_targets)[msk]

        logger.info(f"Reading files.")
        pool = Pool(processes=processes)

        results = pool.starmap(
            read_file,
            zip(
                files_to_read,
                group_valid_targets,
            ),
        )

        values = np.asarray(results)

        properties = dict(
            B_WAVELENGTH=values[0][0],
            R_WAVELENGTH=values[0][4],
            Z_WAVELENGTH=values[0][8],
            B_FLUX=np.vstack(values.T[1]),
            R_FLUX=np.vstack(values.T[5]),
            Z_FLUX=np.vstack(values.T[9]),
            B_IVAR=np.vstack(values.T[2]),
            R_IVAR=np.vstack(values.T[6]),
            Z_IVAR=np.vstack(values.T[10]),
            B_MASK=np.vstack(values.T[3]),
            R_MASK=np.vstack(values.T[7]),
            Z_MASK=np.vstack(values.T[11]),
            TARGETID=np.concatenate(values.T[12]),
        )

        return properties


def main(args: Optional[argparse.Namespace] = None) -> None:
    if args is None:
        args = getArgs()

    level = logging.getLevelName(args.log_level)
    logging.basicConfig(
        stream=sys.stdout,
        level=level,
        format="%(levelname)s:%(message)s",
    )

    if not args.output_prefix.parent.is_dir():
        raise ValueError(f"Invalid output-prefix: {args.output_prefix}")

    data_reader = ReadDESILyaData(
        args.input_directory,
        args.catalogue,
    )

    properties = data_reader.read_data(
        args.downsampling,
        processes=args.num_proc,
    )

    Plots.histogram_qsos_per_masked_bins(
        properties,
        args.output_prefix,
        downsampling=args.downsampling,
        save_data=args.save_data,
        save_plot=True,
        save_dict=vars(args),
    )
    Plots.masked_pixels_percentage(
        properties,
        args.output_prefix,
        downsampling=args.downsampling,
        save_data=args.save_data,
        save_plot=True,
        save_dict=vars(args),
    )


def getArgs() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--input-directory",
        type=Path,
        required=True,
        help="Input directory",
    )

    parser.add_argument(
        "--catalogue",
        type=Path,
        required=True,
        help="QSO catalog",
    )

    parser.add_argument(
        "--num-proc",
        type=int,
        default=None,
        help="Number of processes to use for the multiprocessing tool.",
    )

    parser.add_argument(
        "--downsampling",
        type=float,
        default=1,
        help="Downsampling to apply to input data.",
    )

    parser.add_argument(
        "--save-data", action="store_true", help="Store data to be plotted in py"
    )

    parser.add_argument(
        "--output-prefix",
        type=Path,
        help="Prefix of the path where to save the plots to.",
    )

    parser.add_argument(
        "--log-level",
        default="INFO",
        choices=["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG", "NOTSET"],
    )

    args = parser.parse_args()

    return args


if __name__ == "__main__":
    main()



"""
    Build new deltas format from original deltas. This new format will allow for faster reading of data.
"""
from __future__ import annotations

import argparse
import itertools
import logging
import sys
from configparser import ConfigParser
from multiprocessing import Pool
from pathlib import Path
from typing import TYPE_CHECKING

import fitsio
import numpy as np

from picca_bookkeeper.utils import find_bins

if TYPE_CHECKING:
    from typing import Dict, Optional, Tuple
logger = logging.getLogger(__name__)


def main(args: Optional[argparse.Namespace] = None) -> None:
    if args is None:
        args = getArgs()

    level = logging.getLevelName(args.log_level)
    logging.basicConfig(
        stream=sys.stdout,
        level=level,
        format="%(levelname)s:%(message)s",
    )

    args.out_dir.mkdir(exist_ok=True, parents=True)

    if args.config_for_flux is not None:
        survey_data = read_survey_data(str(args.config_for_flux))
    else:
        survey_data = None

    config = ConfigParser()
    config.read(args.config_file)

    pool = Pool(processes=args.nproc)
    results = pool.starmap(
        convert_delta_file,
        zip(
            args.in_dir.glob("delta*fits*"),
            itertools.repeat(args.out_dir),
            itertools.repeat(config),
            itertools.repeat(survey_data),
        ),
    )
    pool.close()


def convert_delta_file(
    delta_path: Path,
    out_dir: Path,
    config: Dict,
    survey_data: Optional[Tuple[np.ndarray, ...]] = None,
) -> None:
    logger.info(f"Reading file: {delta_path.name}")
    if config["data"]["wave solution"] == "lin":
        lambda_grid = np.arange(
            float(config["data"]["lambda min"]),
            float(config["data"]["lambda max"]) + float(config["data"]["delta lambda"]),
            float(config["data"]["delta lambda"]),
        )
    else:
        lambda_grid = np.arange(
            np.log10(float(config["data"]["lambda min"])),
            np.log10(float(config["data"]["lambda max"]))
            + float(config["data"]["delta log lambda"]),
            float(config["data"]["delta log lambda"]),
        )
        lambda_grid = 10**lambda_grid

    original_delta = fitsio.FITS(delta_path)

    num_forests = len(original_delta) - 1
    num_lambda_bins = len(lambda_grid)

    delta = np.zeros((num_forests, num_lambda_bins), dtype=float)
    weight = np.zeros((num_forests, num_lambda_bins), dtype=float)
    cont = np.zeros((num_forests, num_lambda_bins), dtype=float)
    ra = np.zeros(num_forests, dtype=float)
    dec = np.zeros(num_forests, dtype=float)
    los_id = np.zeros(num_forests, dtype=int)
    z = np.zeros(num_forests, dtype=float)
    meansnr = np.zeros(num_forests, dtype=float)
    tid = np.zeros(num_forests, dtype=int)
    night = np.zeros(num_forests, dtype=str)
    petal = np.zeros(num_forests, dtype=str)
    tile = np.zeros(num_forests, dtype=str)

    if "DELTA_BLIND" in original_delta[1].get_colnames():
        blind = True
    else:
        blind = False

    for i, hdu in enumerate(original_delta[1:]):
        bins = find_bins(
            hdu["LAMBDA"][:],
            lambda_grid,
        )

        if blind:
            delta[i][bins] = hdu["DELTA_BLIND"][:]
        else:
            delta[i][bins] = hdu["DELTA"][:]

        weight[i][bins] = hdu["WEIGHT"][:]
        cont[i][bins] = hdu["CONT"][:]

        header = hdu.read_header()

        ra[i] = header["RA"]
        dec[i] = header["DEC"]
        z[i] = header["Z"]
        los_id[i] = header["LOS_ID"]
        meansnr[i] = header["MEANSNR"]
        tid[i] = header["TARGETID"]
        night[i] = header["NIGHT"]
        petal[i] = header["PETAL"]
        tile[i] = header["TILE"]

    if survey_data is not None:
        raw_ids, raw_flux, raw_ivar, raw_lambda = survey_data
        # Obtain flux information, this only happens if survey
        # data is provided.
        flux = np.full_like(delta, np.nan)
        flux_ivar = np.full_like(delta, np.nan)

        for i, id_ in enumerate(los_id):
            indx = np.searchsorted(raw_ids, id_)

            if indx:
                bins = find_bins(
                    raw_lambda[indx],
                    lambda_grid,
                )

                flux[i][bins] = raw_flux[indx]
                flux_ivar[i][bins] = raw_ivar[indx]
            else:
                print("Forest not found in survey data: ", id_)

    new_header = {
        "WAVE_SOLUTION": header["WAVE_SOLUTION"],
        "DELTA_LAMBDA": header["DELTA_LAMBDA"],
    }
    if blind:
        new_header["BLINDING"] = header["BLINDING"]

    with fitsio.FITS(out_dir / delta_path.name, "rw", clobber=True) as results:
        logger.info(f"Writting file: {delta_path.name}")
        results.write(
            [ra, dec, z, los_id, meansnr, tid, night, petal, tile],
            names=[
                "RA",
                "DEC",
                "Z",
                "LOS_ID",
                "MEANSNR",
                "TARGETID",
                "NIGHT",
                "PETAL",
                "TILE",
            ],
            header=new_header,
            extname="METADATA",
        )

        results.write(
            lambda_grid,
            extname="LAMBDA",
        )

        results.write(
            delta,
            extname="DELTA_BLIND" if blind else "DELTA",
        )

        results.write(
            weight,
            extname="WEIGHT",
        )

        results.write(
            cont,
            extname="CONT",
        )

        if survey_data is not None:
            results.write(flux, extname="FLUX")

            results.write(
                flux_ivar,
                extname="FLUX_IVAR",
            )


def read_survey_data(config_file: Path | str) -> Tuple[np.ndarray, ...]:
    from picca.delta_extraction.survey import Survey

    survey = Survey()
    survey.load_config(str(config_file))
    survey.read_corrections()
    survey.read_masks()
    survey.read_data()
    survey.apply_corrections()
    survey.apply_masks()
    survey.filter_forests()

    ids = np.asarray([forest.los_id for forest in survey.data.forests])
    fluxes = np.asarray([forest.flux for forest in survey.data.forests])
    ivars = np.asarray([forest.ivar for forest in survey.data.forests])
    lambdas = np.asarray([10**forest.log_lambda for forest in survey.data.forests])

    sortinds = ids.argsort()
    return ids[sortinds], fluxes[sortinds], ivars[sortinds], lambdas[sortinds]


def getArgs() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--log-level",
        default="INFO",
        choices=["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG", "NOTSET"],
    )

    parser.add_argument(
        "--in-dir",
        type=Path,
        help="Input directory to read deltas from.",
    )

    parser.add_argument(
        "--out-dir", type=Path, help="Output directory to save deltas to."
    )

    parser.add_argument(
        "--config-file",
        type=Path,
        help="picca delta extraction config file (could be output config file .config.ini).",
    )

    parser.add_argument(
        "--nproc",
        type=int,
        help="Number of processes to use.",
    )

    parser.add_argument(
        "--config-for-flux",
        type=Path,
        help="Compute flux properties as flux and flux variance and add them to delta files. Should use an input picca configuration.",
    )

    args = parser.parse_args()
    return args


if __name__ == "__main__":
    main()



"""Simple script to print default values"""

from __future__ import annotations

from importlib_resources import files

from picca_bookkeeper import resources


def main() -> None:
    print(files(resources).joinpath("example_config.yaml").read_text())


if __name__ == "__main__":
    main()



"""Script to easily compute zeff for a given set of files"""

from __future__ import annotations

import argparse
import logging
import sys
from pathlib import Path
from typing import TYPE_CHECKING

from picca_bookkeeper.utils import compute_zeff

if TYPE_CHECKING:
    from typing import Optional

logger = logging.getLogger(__name__)


def main(args: Optional[argparse.Namespace] = None) -> None:
    if args is None:
        args = getArgs()

    level = logging.getLevelName(args.log_level)
    logging.basicConfig(
        stream=sys.stdout,
        level=level,
        format="%(levelname)s:%(message)s",
    )

    file_names = "\n\t".join(map(str, args.export_files))

    logger.info(f"Files:\n\t{file_names}")

    zeff = compute_zeff(
        export_files=args.export_files,
        rmins=args.rmins,
        rmaxs=args.rmaxs,
    )

    print("Computed zeff: ", zeff)


def getArgs() -> argparse.Namespace:
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "export_files",
        type=Path,
        nargs="*",
        default=[],
    )

    parser.add_argument(
        "--rmins",
        type=float,
        nargs="+",
        default=0,
        help="Minimum distances.",
    )

    parser.add_argument(
        "--rmaxs",
        type=float,
        nargs="+",
        default=300,
        help="Maximum distances.",
    )

    parser.add_argument(
        "--log-level",
        default="INFO",
        choices=["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG", "NOTSET"],
    )

    args = parser.parse_args()
    return args


if __name__ == "__main__":
    main()



""" Read all masks applied during a picca run, for studying its effects"""

from __future__ import annotations

import argparse
import itertools
import logging
import multiprocessing
import sys
from pathlib import Path
from typing import TYPE_CHECKING

import fitsio
import numpy as np
from picca.delta_extraction.astronomical_objects.forest import Forest
from picca.delta_extraction.masks.bal_mask import BalMask
from picca.delta_extraction.masks.dla_mask import DlaMask, dla_profile
from picca.delta_extraction.survey import Survey

from picca_bookkeeper.utils import find_bins

if TYPE_CHECKING:
    from typing import Dict, List, Optional

logger = logging.getLogger(__name__)


class ReadExtraDeltasData:
    """Class to read extra data from picca deltas."""

    def __init__(
        self,
        picca_config: Path | str,
    ):
        """Initialize class instance

        Args:
            picca_config: Picca configuration file (output is better since it has all the info.)
        """
        self.picca_config = picca_config

        self.survey = Survey()
        self.survey.load_config(str(self.picca_config))

    def _get_survey_mask(self, classinfo: DlaMask | BalMask) -> DlaMask | BalMask:
        for mask in self.survey.masks:
            if isinstance(mask, classinfo):
                return mask
        else:
            raise ValueError(
                "Could not find compatible mask in survey masks", self.survey.masks
            )

    def read_data(self) -> None:
        """Read data from DESI using picca survey object."""
        self.survey.read_corrections()
        self.survey.read_masks()
        self.survey.read_data()
        self.survey.apply_corrections()

    def read_DLA_mask(self, num_processors: int) -> None:
        """Read DLA mask using picca code."""
        logger.info("Reading DLA masks.")
        mask = self._get_survey_mask(DlaMask)

        context = multiprocessing.get_context("fork")
        chunksize = int(len(self.survey.data.forests) / num_processors / 3)
        chunksize = max(1, chunksize)

        with context.Pool(processes=num_processors) as pool:
            self.survey.data.forests = pool.starmap(
                read_DLA_mask_single,
                zip(
                    self.survey.data.forests,
                    itertools.repeat(mask),
                ),
                chunksize=chunksize,
            )

    def read_BAL_mask(self, num_processors: int) -> None:
        """Read BAL mask using picca code."""
        logger.info("Reading BAL masks.")
        mask = self._get_survey_mask(BalMask)

        context = multiprocessing.get_context("fork")
        chunksize = int(len(self.survey.data.forests) / num_processors / 3)
        chunksize = max(1, chunksize)

        # for forest in self.survey.data.forests:
        #     read_BAL_mask_single(forest, mask)

        with context.Pool(processes=num_processors) as pool:
            self.survey.data.forests = pool.starmap(
                read_BAL_mask_single,
                zip(
                    self.survey.data.forests,
                    itertools.repeat(mask),
                ),
                chunksize=chunksize,
            )

    def save_data_mask(
        self,
        out_dir: Path | str,
        sorted_ids: Dict,
        num_processors: Optional[int] = None,
    ) -> None:
        logger.info("Converting statistics into rest frame.")
        Path(out_dir).mkdir(exist_ok=True)

        context = multiprocessing.get_context("fork")
        args = [(forest,) for forest in self.survey.data.forests]
        with context.Pool(processes=num_processors) as pool:
            forests = pool.starmap(
                compute_RF_statistics,
                args,
            )

        logger.info("Saving mask data.")
        healpixs = np.array([forest.healpix for forest in forests])
        unique_healpixs = np.unique(healpixs)

        arguments = []
        for healpix in unique_healpixs:
            this_idx = np.nonzero(healpix == healpixs)[0]
            grouped_forests = [forests[i] for i in this_idx]
            arguments.append((out_dir, healpix, grouped_forests, sorted_ids[healpix]))

        context = multiprocessing.get_context("fork")
        with context.Pool(processes=num_processors) as pool:
            pool.starmap(
                save_healpix_data_mask,
                arguments,
            )

    def save_data_flux(
        self,
        out_dir: Path | str,
        sorted_ids: Dict,
        num_processors: Optional[int] = None,
    ) -> None:
        Path(out_dir).mkdir(exist_ok=True)

        logger.info("Saving flux data.")
        forests = self.survey.data.forests
        healpixs = np.array([forest.healpix for forest in forests])
        unique_healpixs = np.unique(healpixs)

        arguments = []
        for healpix in unique_healpixs:
            this_idx = np.nonzero(healpix == healpixs)[0]
            grouped_forests = [forests[i] for i in this_idx]
            arguments.append((out_dir, healpix, grouped_forests, sorted_ids[healpix]))

        context = multiprocessing.get_context("fork")
        with context.Pool(processes=num_processors) as pool:
            pool.starmap(
                save_healpix_data_flux,
                arguments,
            )

    def read_original_deltas_ids(self, in_dir: Path | str) -> Dict:
        logger.info("Reading original LOS_ID sorting.")
        target_ids = dict()
        for delta_file in Path(in_dir).glob("delta-*.fits.gz"):
            healpix = int(delta_file.name.split("-")[1].split(".")[0])
            with fitsio.FITS(delta_file) as hdul:
                target_ids[healpix] = hdul["METADATA"].read()["LOS_ID"]

        return target_ids


def compute_RF_statistics(
    forest: Forest,
) -> Forest:
    bins = find_bins(
        10**forest.log_lambda / (1 + forest.z),
        10**Forest.log_lambda_rest_frame_grid,
    )

    count = np.bincount(bins, minlength=len(Forest.log_lambda_rest_frame_grid))
    w = count > 0

    rebin = np.bincount(
        bins, weights=forest.dla_mask, minlength=len(Forest.log_lambda_rest_frame_grid)
    )
    forest.dla_mask_rf = np.array(rebin, dtype=float)
    forest.dla_mask_rf[w] /= count[w]

    rebin = np.bincount(
        bins, weights=forest.bal_mask, minlength=len(Forest.log_lambda_rest_frame_grid)
    )
    forest.bal_mask_rf = np.array(rebin, dtype=float)
    forest.bal_mask_rf[w] /= count[w]

    rebin = np.bincount(
        bins,
        weights=forest.dla_transmission,
        minlength=len(Forest.log_lambda_rest_frame_grid),
    )
    forest.dla_transmission_rf = np.array(rebin, dtype=float)
    forest.dla_transmission_rf[w] /= count[w]

    forest.log_lambda_rest_frame_index = np.unique(bins)

    return forest


def save_healpix_data_mask(
    out_dir: Path | str,
    healpix: int,
    forests: List[Forest],
    sorted_ids: List[int],
) -> None:
    out_dir = Path(out_dir)

    # Sorting forests by LOS_ID will make identification faster
    forests = sorted(forests, key=lambda x: x.los_id)
    ids = [forest.los_id for forest in forests]

    # Check all original LOS_IDs are present
    if not np.all(np.in1d(sorted_ids, ids)):
        logger.warning("Input deltas have LOS_IDS that do not appear on extended data")

    # We search each sorted_ids into ids (which is sorted numerically to speed up the process)
    save_indexs = np.searchsorted(ids, sorted_ids)
    save_forests = np.asarray(forests)[save_indexs]

    with fitsio.FITS(
        out_dir / f"mask-info-{healpix}.fits.gz", "rw", clobber=True
    ) as hdul:
        hdul.write(
            np.array(
                [tuple(forest.get_metadata()) for forest in save_forests],
                dtype=save_forests[0].get_metadata_dtype(),
            ),
            extname="METADATA",
        )

        hdul.write(
            10**Forest.log_lambda_grid,
            extname="LAMBDA",
        )

        hdul.write(
            10**Forest.log_lambda_rest_frame_grid,
            extname="LAMBDA_RF",
        )

        in_mask = np.full((len(save_forests), len(Forest.log_lambda_grid)), 0)
        for i, forest in enumerate(save_forests):
            in_mask[i][forest.log_lambda_index] = np.ones_like(
                forest.log_lambda_index, dtype=int
            )
        hdul.write(
            in_mask,
            extname="INPUT_MASK",
        )

        dla_mask = np.full((len(save_forests), len(Forest.log_lambda_grid)), False)
        for i, forest in enumerate(save_forests):
            dla_mask[i][forest.log_lambda_index] = forest.dla_mask
        hdul.write(
            np.array(dla_mask, dtype=int),
            extname="DLA_MASK",
        )

        dla_transmission = np.full(
            (len(save_forests), len(Forest.log_lambda_grid)),
            0,
            dtype=float,
        )
        for i, forest in enumerate(save_forests):
            dla_transmission[i][forest.log_lambda_index] = forest.dla_transmission
        hdul.write(dla_transmission, extname="DLA_TRANSMISSION")

        bal_mask = np.full((len(save_forests), len(Forest.log_lambda_grid)), False)
        for i, forest in enumerate(save_forests):
            bal_mask[i][forest.log_lambda_index] = forest.bal_mask
        hdul.write(np.array(bal_mask, dtype=int), extname="BAL_MASK")

        # SAVE RF properties
        hdul.write(
            10**Forest.log_lambda_rest_frame_grid,
            extname="LAMBDA_RF",
        )

        in_mask = np.full(
            (len(save_forests), len(Forest.log_lambda_rest_frame_grid)), 0
        )
        for i, forest in enumerate(save_forests):
            in_mask[i][forest.log_lambda_rest_frame_index] = np.ones_like(
                forest.log_lambda_rest_frame_index, dtype=int
            )
        hdul.write(
            in_mask,
            extname="INPUT_MASK_RF",
        )

        dla_mask_rf = np.asarray([forest.dla_mask_rf for forest in save_forests])
        hdul.write(
            np.array(dla_mask_rf, dtype=int),
            extname="DLA_MASK_RF",
        )

        dla_transmission_rf = np.asarray(
            [forest.dla_transmission_rf for forest in save_forests]
        )
        hdul.write(dla_transmission_rf, extname="DLA_TRANSMISSION_RF")

        bal_mask_rf = np.asarray([forest.bal_mask_rf for forest in save_forests])
        hdul.write(np.array(bal_mask_rf, dtype=int), extname="BAL_MASK_RF")


def save_healpix_data_flux(
    out_dir: Path | str,
    healpix: int,
    forests: List[Forest],
    sorted_ids: List[int],
) -> None:
    out_dir = Path(out_dir)

    # Sorting forests by LOS_ID will make identification faster
    forests = sorted(forests, key=lambda x: x.los_id)
    ids = [forest.los_id for forest in forests]

    # Check all original LOS_IDs are present
    if not np.all(np.in1d(sorted_ids, ids)):
        logger.warning("Input deltas have LOS_IDS that do not appear on extended data")

    # We search each sorted_ids into ids (which is sorted numerically to speed up the process)
    save_indexs = np.searchsorted(ids, sorted_ids)
    save_forests = np.asarray(forests)[save_indexs]

    with fitsio.FITS(
        out_dir / f"flux-info-{healpix}.fits.gz", "rw", clobber=True
    ) as hdul:
        hdul.write(
            np.array(
                [tuple(forest.get_metadata()) for forest in save_forests],
                dtype=save_forests[0].get_metadata_dtype(),
            ),
            extname="METADATA",
        )

        flux = np.full((len(save_forests), len(Forest.log_lambda_grid)), np.nan)
        for i, forest in enumerate(save_forests):
            flux[i][forest.log_lambda_index] = forest.flux

        hdul.write(
            flux,
            extname="FLUX",
        )

        ivar = np.full((len(save_forests), len(Forest.log_lambda_grid)), np.nan)
        for i, forest in enumerate(save_forests):
            ivar[i][forest.log_lambda_index] = forest.ivar

        hdul.write(
            ivar,
            extname="FLUX_IVAR",
        )


def read_DLA_mask_single(forest: Forest, mask: DlaMask) -> Forest:
    lambda_ = 10**forest.log_lambda

    # load DLAs
    if mask.los_ids.get(forest.los_id) is not None:
        dla_transmission = np.ones(len(lambda_))
        for z_abs, nhi in mask.los_ids.get(forest.los_id):
            dla_transmission *= dla_profile(lambda_, z_abs, nhi)

        # find out which pixels to mask
        w = dla_transmission > mask.dla_mask_limit
        if len(mask.mask) > 0:
            for mask_range in mask.mask:
                for z_abs, nhi in mask.los_ids.get(forest.los_id):
                    w &= (lambda_ / (1.0 + z_abs) < mask_range["wave_min"]) | (
                        lambda_ / (1.0 + z_abs) > mask_range["wave_max"]
                    )

        # do the actual masking
        forest.transmission_correction *= dla_transmission
        forest.dla_transmission = dla_transmission
        forest.dla_mask = w
    else:
        forest.dla_mask = np.ones_like(lambda_)
        forest.dla_transmission = forest.transmission_correction

    return forest


def read_BAL_mask_single(forest: Forest, mask: BalMask) -> Forest:
    if not hasattr(forest, "los_id"):
        pass
    if not hasattr(mask, "los_ids"):
        pass

    mask_table = mask.los_ids.get(forest.los_id)
    if (mask_table is None) or len(mask_table) == 0:
        forest.bal_mask = np.ones_like(forest.log_lambda)
        return forest

    # find out which pixels to mask
    w = np.ones(forest.log_lambda.size, dtype=bool)
    rest_frame_log_lambda = forest.log_lambda - np.log10(1.0 + forest.z)
    mask_idx_ranges = np.searchsorted(
        rest_frame_log_lambda,
        [mask_table["log_lambda_min"], mask_table["log_lambda_max"]],
    ).T
    # Make sure first index comes before the second
    mask_idx_ranges.sort(axis=1)

    for idx1, idx2 in mask_idx_ranges:
        w[idx1:idx2] = 0

    forest.bal_mask = w

    return forest


def main(args: Optional[argparse.Namespace] = None) -> None:
    if args is None:
        args = getArgs()

    level = logging.getLevelName(args.log_level)
    logging.basicConfig(
        stream=sys.stdout,
        level=level,
        format="%(levelname)s:%(message)s",
    )

    read = ReadExtraDeltasData(args.picca_config)
    read.read_data()

    target_ids = read.read_original_deltas_ids(in_dir=args.deltas_input)
    if args.output_dir_mask is not None:
        read.read_DLA_mask(args.num_processors)
        read.read_BAL_mask(args.num_processors)
        read.save_data_mask(
            out_dir=args.output_dir_mask,
            sorted_ids=target_ids,
            num_processors=args.num_processors,
        )

    if args.output_dir_flux is not None:
        read.survey.apply_masks()
        read.survey.filter_forests()

        read.save_data_flux(
            out_dir=args.output_dir_flux,
            sorted_ids=target_ids,
            num_processors=args.num_processors,
        )
    logger.info("Done.")


def getArgs() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--deltas-input",
        type=Path,
        required=True,
        help="Deltas to add information to (used for sorting)",
    )
    parser.add_argument(
        "--picca-config",
        type=Path,
        required=True,
        help="picca configuration file (output is better since it has all the info.)",
    )

    parser.add_argument(
        "--num-processors",
        type=int,
        help="num of processors to use with the multiprocessing tool.",
    )

    parser.add_argument(
        "--output-dir-mask",
        type=Path,
        required=False,
        default=None,
        help="Output directory to save mask info into.",
    )

    parser.add_argument(
        "--output-dir-flux",
        type=Path,
        required=False,
        default=None,
        help="Output directory to save flux info into.",
    )

    parser.add_argument(
        "--log-level",
        default="INFO",
        choices=["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG", "NOTSET"],
    )

    args = parser.parse_args()
    return args


if __name__ == "__main__":
    main()



"""
    Script to combine multiple DLA catalogs.
"""

from __future__ import annotations

import argparse
import logging
import sys
import textwrap
from pathlib import Path
from typing import TYPE_CHECKING

import astropy
import fitsio
import numpy as np
from astropy.table import Table, unique, vstack

if TYPE_CHECKING:
    from typing import Dict, Optional

logger = logging.getLogger(__name__)


def select_table(
    table: astropy.table.table.Table,
    selection: int = 2,
    NHI: float = 20.3,
    S2N: float = 0,
) -> astropy.table.table.Table:
    """Select on confidence flags/snr, also select type==DLA only"""
    # print('first select unique DLA ID')
    # tab = select_table_1(tab)

    logger.info("Start selecting on conf./SNR")

    if selection == 2:
        if "CNN_DLA_CONFIDENCE" in table.keys() and "GP_DLA_CONFIDENCE" in table.keys():
            conf_min = np.minimum(
                table["CNN_DLA_CONFIDENCE"],
                table["GP_DLA_CONFIDENCE"],
            )
        else:
            conf_min = table["DLA_CONFIDENCE"]

        msk = table["ABSORBER_TYPE"] == "DLA"
        msk &= conf_min > 0.5
    elif selection == 3:
        conf_cnn = table["CNN_DLA_CONFIDENCE"]
        msk_cnn = ((conf_cnn > 0.3) & (table["S2N"] > 0)) | (
            (conf_cnn > 0.2) & (table["S2N"] >= 3)
        )

        msk_gp = table["GP_DLA_CONFIDENCE"] > 0.9

        msk = table["ABSORBER_TYPE"] == "DLA"
        msk &= msk_gp | msk_cnn
    elif selection == 4:
        conf = table["CNN_DLA_CONFIDENCE"]

        msk = table["ABSORBER_TYPE"] == "DLA"
        msk &= ((conf > 0.3) & (table["S2N"] > 0)) | (
            (conf > 0.2) & (table["S2N"] >= 3)
        )
    elif selection == 5:
        conf_min = np.minimum(table["CNN_DLA_CONFIDENCE"], table["GP_DLA_CONFIDENCE"])
        msk = table["NHI"] > NHI
        msk &= conf_min > 0.5
        msk &= table["S2N"] > S2N
    elif selection == 6:
        # gp_nhi being 0 means nhi is taken from cnn
        # we try correcting the cnn_nhi by adding 0.17
        idx = np.where(table["GP_NHI"] == 0)
        table["NHI"][idx] += 0.17

        msk = table["NHI"] > NHI
        msk &= table["CNN_DLA_CONFIDENCE"] > 0.5
        msk &= table["S2N"] > S2N

    else:
        raise ValueError("Selection not valid.")

    return table[msk]


def main(args: Optional[argparse.Namespace] = None) -> None:
    if args is None:
        args = getArgs()

    level = logging.getLevelName(args.log_level)
    logging.basicConfig(
        stream=sys.stdout,
        level=level,
        format="%(levelname)s:%(message)s",
    )

    logger.info("Reading DLA catalogues.")
    catalogues = [
        Table(fitsio.FITS(catalogue)[1].read()) for catalogue in args.dla_catalogues
    ]

    logger.info("Selecting objects from catalogues.")
    if args.selection != 0:
        catalogues = [
            select_table(
                catalogue, selection=args.selection, NHI=args.NHI, S2N=args.S2N
            )
            for catalogue in catalogues
        ]

    logger.info("Combining catalogues.")
    final_catalogue = vstack(catalogues)
    final_catalogue = unique(final_catalogue)
    columns = list(final_catalogue.columns)

    logger.info("Writing final catalogue.")
    with fitsio.FITS(args.output_catalogue, "rw", clobber=True) as results:
        header: Dict = {}
        results.write(
            [final_catalogue[column] for column in columns],
            names=columns,
            header=header,
            extname="DLACAT",
        )
    logger.info("Done.")


def getArgs() -> argparse.Namespace:
    parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument(
        "dla_catalogues",
        type=Path,
        nargs="+",
        help="Input DLA catalogs",
    )

    parser.add_argument(
        "--output-catalogue",
        type=Path,
        required=True,
        help="Output path to the final DLA catalogue.",
    )

    parser.add_argument(
        "--NHI",
        type=float,
        default=20.3,
        required=False,
        help="minimum DLA NHI, used only in selection 5 or 6",
    )

    parser.add_argument(
        "--S2N",
        type=float,
        default=0,
        required=False,
        help="minimum QSO S2N, used only in selection 5 or 6",
    )

    parser.add_argument(
        "--selection",
        type=int,
        default=0,
        choices=[
            0,
            2,
            3,
            4,
            5,
            6,
        ],
        help=textwrap.dedent(
            """
0: No cuts.
1: Invalid (to reserve v1 for full catalog.)
2: (IFAE suggestion). Built to generate approximately 1 DLA for each 3 QSOs.
    Select DLA absorber type.
    Select absorbers detected by both CNN and GP with confidence>0.5.
3: DLA paper recommendation.
    Select DLA absorber type
    Select absorbers fulfilling any:
        - For CNN: select absorbers with 'DLA_CONFIDENCE'>0.2 as valid detections for 'S2N'>3, 'DLA_CONFIDENCE'>0.3 for 'S2N'<3.
        - For GP: select absorbers with ¡DLA_CONFIDENCE'>0.9.
4: Original recommendation by DLA Team.
    Select DLA absorber type.
    Select absorbers with 'DLA_CONFIDENCE'>0.2 as valid detections for 'S2N'>3, 'DLA_CONFIDENCE'>0.3 for 'S2N'<3.
    Only use CNN.
5: CONF_CNN > 0.5 and CONF_GP > 0.5 and NHI > args.NHI and S2N > args.S2N
6: CONF_CNN > 0.5, if NHI_GP available use else use NHI_CNN + 0.17, NHI > args.NHI and S2N > args.S2N
            """
        ),
    )

    parser.add_argument(
        "--log-level",
        default="INFO",
        choices=["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG", "NOTSET"],
    )

    args = parser.parse_args()
    return args


if __name__ == "__main__":
    main()



"""
Script to modify all the links/paths inside the bookkeeper to match
the current path structure.
"""

from __future__ import annotations

import argparse
import logging
import re
import sys
from pathlib import Path
from typing import TYPE_CHECKING

import yaml

if TYPE_CHECKING:
    from typing import Optional

logger = logging.getLogger(__name__)


def main(args: Optional[argparse.Namespace] = None) -> None:
    if args is None:
        args = getArgs()

    level = logging.getLevelName(args.log_level)
    logging.basicConfig(
        stream=sys.stdout,
        level=level,
        format="%(levelname)s:%(message)s",
    )

    old_dirs = set(args.replace_dirs)

    if not args.only_replace_parsed_dirs:
        yaml_files = list(args.source_dir.glob("**/bookkeeper_config.yaml"))
        for file in yaml_files:
            config = yaml.safe_load(file.read_text())

            old_dirs.add(str(Path(config["data"]["bookkeeper dir"]).parent))

    yaml_files = list(args.source_dir.glob("**/*.yaml"))

    file_string = "\n\t".join(map(str, yaml_files))
    print(f"Config files to be modified:\n\t{file_string}")

    slurm_files = list(args.source_dir.glob("**/*.sh"))
    file_string = "\n\t".join(map(str, slurm_files))
    print(f"Slurm files to be modified:\n\t{file_string}")

    ini_files = list(args.source_dir.glob("**/*.ini"))
    file_string = "\n\t".join(map(str, ini_files))
    print(f"Ini files to be modified:\n\t{file_string}")

    links = list(args.source_dir.glob("**/*.fits"))
    links += list(args.source_dir.glob("**/*.fits.gz"))
    links = [x for x in links if x.is_symlink()]

    file_string = "\n\t".join(map(str, links))
    print(f"Symlink files:\n\t{file_string}")

    # Remove ending / because if there are 2 issues might happen
    updated_old_dirs = []
    for old_dir in old_dirs:
        while old_dir[-1] == "/":
            old_dir = old_dir[:-1]
        updated_old_dirs.append(old_dir)
    old_dirs = set(updated_old_dirs)

    new_dir = args.new_dir
    while new_dir[-1] == "/":
        new_dir = new_dir[:-1]
    print(f"New directory:\n\t{new_dir}")

    print("\nThe following substitutions will be made:")
    for dir_ in old_dirs:
        print(strRed(dir_), "->", strCyan(new_dir))

    answer = input(
        "This action cannot be undone, are you really sure? Please write 'Yes, I am' to proceed:\n"
    )

    if answer != "Yes, I am":
        print("Exiting")
        exit()

    logger.info("Rewriting config files.")
    for yaml_file in yaml_files:
        for dir_ in old_dirs:
            with open(yaml_file, "r") as file:
                lines = file.readlines()

            with open(yaml_file, "w") as file:
                for line in lines:
                    file.write(re.sub(dir_, new_dir, line))

    logger.info("Rewriting slurm files.")
    for slurm_file in slurm_files:
        for dir_ in old_dirs:
            with open(slurm_file, "r") as file:
                lines = file.readlines()

            with open(slurm_file, "w") as file:
                for line in lines:
                    file.write(re.sub(dir_, new_dir, line))

    logger.info("Rewriting ini files.")
    for ini_file in ini_files:
        for dir_ in old_dirs:
            with open(ini_file, "r") as file:
                lines = file.readlines()

            with open(ini_file, "w") as file:
                for line in lines:
                    file.write(re.sub(dir_, new_dir, line))

    logger.info("Modifying symbolic links.")
    for link in links:
        for dir_ in old_dirs:
            if dir_ in str(link.readlink().absolute()):
                new_link = Path(re.sub(dir_, new_dir, str(link.readlink().absolute())))

                link.unlink()
                link.symlink_to(new_link)


def strRed(skk: str) -> str:
    return "\033[91m {}\033[00m".format(skk)


def strCyan(skk: str) -> str:
    return "\033[34m {}\033[00m".format(skk)


def getArgs() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawTextHelpFormatter,
        description="""
Script to modify all the links/paths inside a bookkeeper structure to match the current path structure where the files are located in. The internal structure of the files shouldn't have been modified.

This script will modify the following things under the given bookkeeper-dir:
    - All .yaml files under the given bookkeeper-dir. To substitute the "bookkeeper dir" specified in the files for the real one.
    - All .sl files, from the collected "bookkeeper dir" in the previous step into the real one.
    - All symlinks, again from the collected "bookkeeper dir" into the real one.
""",
    )

    parser.add_argument(
        "source_dir",
        type=Path,
        help="Directory where to search for bookkeeper_config.yaml files "
        "to start the process.",
    )

    parser.add_argument(
        "--new-dir",
        default=None,
        type=str,
        help="New bookkeeper dir that should be written in all files.",
    )

    parser.add_argument(
        "--replace-dirs",
        default=[],
        type=str,
        nargs="+",
        help="Add extra string to be substituted for the new-dir.",
    )

    parser.add_argument(
        "--only-replace-parsed-dirs",
        action="store_true",
        help="Only replace directories that have been parsed (not the ones in current structure).",
    )

    parser.add_argument(
        "--log-level",
        default="INFO",
        choices=["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG", "NOTSET"],
    )

    args = parser.parse_args()
    return args


if __name__ == "__main__":
    main()



""" Script to run vega sampler given a bookkeeper config file"""

from __future__ import annotations

import argparse
import logging
import sys
from pathlib import Path
from typing import TYPE_CHECKING

from picca_bookkeeper.bookkeeper import Bookkeeper
from picca_bookkeeper.dict_utils import DictUtils

if TYPE_CHECKING:
    from typing import Optional
logger = logging.getLogger(__name__)


def main(args: Optional[argparse.Namespace] = None) -> None:
    if args is None:
        args = get_args()

    level = logging.getLevelName(args.log_level)
    logging.basicConfig(
        stream=sys.stdout,
        level=level,
        format="%(levelname)s:%(message)s",
    )

    logger.info("Adding sampler.")

    bookkeeper = Bookkeeper(
        args.bookkeeper_config,
        overwrite_config=args.overwrite_config,
        read_mode=False,
    )

    config = DictUtils.merge_dicts(
        bookkeeper.defaults,
        bookkeeper.config,
    )

    if config["fits"].get("compute covariance", False):
        compute_covariance = bookkeeper.get_covariance_matrix_tasker(
            wait_for=args.wait_for,
            system=args.system,
            overwrite=args.overwrite,
            skip_sent=args.skip_sent,
        )
        compute_covariance.write_job()
        if not args.only_write:
            compute_covariance.send_job()
            logger.info(f"Sent compute covariance:\n\t{compute_covariance.jobid}")

        if config["fits"].get("smooth covariance", False):
            smooth_covariance = bookkeeper.get_smooth_covariance_tasker(
                wait_for=args.wait_for,
                system=args.system,
                overwrite=args.overwrite,
                skip_sent=args.skip_sent,
            )
            smooth_covariance.write_job()
            if not args.only_write:
                smooth_covariance.send_job()
                logger.info(f"Sent smooth covariance:\n\t{smooth_covariance.jobid}")

    if config["fits"].get("compute zeff", False):
        compute_zeff = bookkeeper.get_compute_zeff_tasker(
            wait_for=args.wait_for,
            overwrite=args.overwrite,
            skip_sent=args.skip_sent,
        )
        compute_zeff.write_job()
        if not args.only_write:
            compute_zeff.send_job()
            logger.info(f"Sent compute zeff:\n\t{compute_zeff.jobid}")

    sampler = bookkeeper.get_sampler_tasker(
        wait_for=args.wait_for,
        overwrite=args.overwrite,
        skip_sent=args.skip_sent,
    )

    sampler.write_job()
    if not args.only_write:
        sampler.send_job()
        logger.info(f"Sent fit:\n\t{sampler.jobid}")


def get_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "bookkeeper_config", type=Path, help="Path to bookkeeper file to use"
    )

    parser.add_argument(
        "--overwrite", action="store_true", help="Force overwrite output data."
    )

    parser.add_argument(
        "--overwrite-config",
        action="store_true",
        help="Force overwrite bookkeeper config.",
    )

    parser.add_argument(
        "--skip-sent", action="store_true", help="Skip runs that were already sent."
    )

    parser.add_argument(
        "--only-write", action="store_true", help="Only write scripts, not send them."
    )

    parser.add_argument("--wait-for", nargs="+", type=int, default=None, required=False)

    parser.add_argument(
        "--log-level",
        default="INFO",
        choices=["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG", "NOTSET"],
    )

    args = parser.parse_args()

    return args


if __name__ == "__main__":
    main()



""" Script to run picca_delta_extraction or picca_convert_transmission
given a bookkeeper config file."""

from __future__ import annotations

import argparse
import logging
import sys
from pathlib import Path
from typing import TYPE_CHECKING

from picca_bookkeeper.bookkeeper import Bookkeeper
from picca_bookkeeper.dict_utils import DictUtils
from picca_bookkeeper.tasker import ChainedTasker, DummyTasker

if TYPE_CHECKING:
    from typing import Callable, Optional, Type
logger = logging.getLogger(__name__)


def main(args: Optional[argparse.Namespace] = None) -> None:
    if args is None:
        args = get_args()

    level = logging.getLevelName(args.log_level)
    logging.basicConfig(
        stream=sys.stdout,
        level=level,
        format="%(levelname)s:%(message)s",
    )

    bookkeeper = Bookkeeper(
        args.bookkeeper_config,
        overwrite_config=args.overwrite_config,
        read_mode=False,
    )

    config = DictUtils.merge_dicts(
        bookkeeper.defaults,
        bookkeeper.config,
    )

    raw = config["general"].get("raw mocks", False)
    true = config["general"].get("true mocks", False)

    if args.only_calibration or (
        (
            (not raw and not true)
            and config["delta extraction"]["calib"] != 0
            and not args.skip_calibration
        )
    ):
        logger.info("Adding calibration step(s).")
        calibration = bookkeeper.get_calibration_extraction_tasker(
            system=args.system,
            debug=args.debug,
            wait_for=args.wait_for,
            overwrite=args.overwrite,
            skip_sent=args.skip_sent,
        )
        calibration.write_job()
        if not args.only_write:
            calibration.send_job()

            if isinstance(calibration, ChainedTasker):
                for tasker in calibration.taskers:
                    if not isinstance(tasker, DummyTasker):
                        logger.info(f"Sent calibration step:\n\t{tasker.jobid}")
            elif not isinstance(calibration, DummyTasker):
                logger.info(f"Sent calibration step:\n\t{calibration.jobid}")

        logger.info("Done.\n")

        if args.only_calibration:
            return

        get_tasker: Callable = bookkeeper.get_delta_extraction_tasker
    elif raw:
        get_tasker = bookkeeper.get_raw_deltas_tasker
    else:
        get_tasker = bookkeeper.get_delta_extraction_tasker

    logger.info(f"Adding deltas for region: {args.region}.")
    deltas = get_tasker(
        region=args.region,
        system=args.system,
        debug=args.debug,
        wait_for=args.wait_for,
        overwrite=args.overwrite,
        skip_sent=args.skip_sent,
    )
    deltas.write_job()
    if not args.only_write:
        deltas.send_job()

        if not isinstance(deltas, DummyTasker):
            logger.info(f"Sent deltas for region:\n\t{args.region}: {deltas.jobid}")

    logger.info("Done\n")


def get_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "bookkeeper_config", type=Path, help="Path to bookkeeper file to use"
    )

    parser.add_argument(
        "--system",
        type=str,
        default=None,
    )

    parser.add_argument(
        "--region",
        type=str,
        default="lya",
        help="Region to compute deltas in",
    )

    parser.add_argument(
        "--overwrite", action="store_true", help="Force overwrite output data."
    )

    parser.add_argument(
        "--overwrite-config",
        action="store_true",
        help="Force overwrite bookkeeper config.",
    )

    parser.add_argument(
        "--skip-sent", action="store_true", help="Skip runs that were already sent."
    )

    parser.add_argument(
        "--debug",
        action="store_true",
    )

    parser.add_argument(
        "--only-calibration",
        action="store_true",
        help="Only compute calibration steps.",
    )

    parser.add_argument(
        "--skip-calibration",
        action="store_true",
        help="Skip calibration step if already computed.",
    )

    parser.add_argument(
        "--only-write",
        action="store_true",
        help="Only write scripts, do not send them.",
    )

    parser.add_argument("--wait-for", nargs="+", type=int, default=None, required=False)

    parser.add_argument(
        "--log-level",
        default="INFO",
        choices=["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG", "NOTSET"],
    )

    args = parser.parse_args()

    return args


if __name__ == "__main__":
    main()
